/**************************************************************************
*                                                                         *
*                             COPYRIGHT NOTICE                            *
*                                                                         *
* This software/database is categorized as "United States Government      *
* Work" under the terms of the United States Copyright Act.  It was       *
* produced as part of the author's official duties as a Government        *
* employee and thus can not be copyrighted.  This software/database is    *
* freely available to the public for use without a copyright notice.      *
* Restrictions can not be placed on its present or future use.            *
*                                                                         *
* Although all reasonable efforts have been taken to ensure the accuracy  *
* and reliability of the software and data, the National Library of       *
* Medicine (NLM) and the U.S. Government do not and can not warrant the   *
* performance or results that may be obtained by using this software,     *
* data, or derivative works thereof.  The NLM and the U.S. Government     *
* disclaim any and all warranties, expressed or implied, as to the        *
* performance, merchantability or fitness for any particular purpose or   *
* use.                                                                    *
*                                                                         *
* In any work or product derived from this material, proper attribution   *
* of the author(s) as the source of the software or data would be         *
* appreciated.                                                            *
*                                                                         *
**************************************************************************/
/*
BLAST3 - Search a protein database for regions that approximately match
regions of a query sequence.  The basic command syntax is

 	BLAST3 database query_sequence

 where:
 	database names a file containing protein sequences, separated by
 		header lines beginning with the character '>'.  A typical
 		sequence entry is:

		>CCHU (PIR) Cytochrome c - Human | 1.0,1.0,1.0,1.0,1.0
		GDVEKGKKIFIMKCSQCHTVEKGGKHKTGP
		NLHGLFGRKTGQAPGYSYTAANKNKGIIWG
		EDTLMEYLENPKKYIPGTKMIFVGIKKKEE
		RADLIAYLKKATNE

 	query_sequence names a file containing a protein sequence (where the
 		header line is optional).
 
 The BLAST3 command permits optional additional arguments, such as X=50,
 that reset certain internal parameters.  The available parameters are:


Parameters controlling 3-way alignment construction:

	F or f is the highest Expect-value of a reported three-way alignment
		(default is 5.)
	L or l is the lowest score permitted in reporting three-way alignments
		(if specified, it overrides F parameter)
	R or r is the maximum Ratio of the Expect-value for a 3-way alignment
		to the smallest Expect-value of its component pairwise alignments
		(the default value is 2, but the user may very well want to set
		this parameter to 1, 0.5, OR LESS, to avoid what might be copious
		"garbage" output)
	N or n is the highest pairwise Expect-value to use when constructing
		3-way alignments (default is 0.1).  In essence, this defines
		which pairwise alignments are insignificant and should be used
		to find significant 3-way alignments.
	U or u is the highest pairwise score to use when constructing 3-way
		alignments (if specified, it overrides P parameter)


Parameters controlling pairwise alignment construction (initial BLAST search):

	E or e is the estimated number of random HSPs to collect during the
		initial BLAST portion of the search (default is 1000).  This is
		close to the number of diagonals remembered for construction
		of three-way alignment.
	S or s gives the High-scoring Segment Pair (HSP) cutoff score for the
		initial BLAST portion of the search (default depends on E parameter)
	T or t gives the threshold for a word match during the initial BLAST
		portion of the search (default is 13 or 12)
	X or x gives the value for terminating word hit extensions during the
		initial BLAST portion of the search (default is 20)


Parameters controlling both pairwise and three-way alignment construction:

	M or m is the name of a file containing substitution costs
		(default is "BLOSUM62")
	Y or y is the effective query sequence length for estimating statistical
		significance (default is the actual length of the query sequence)
	Z or z is the effective database length for estimating statistical
		significance (default is the actual length of the database)


Miscellaneous parameters:

	P or p is the number of processors to utilize when the program
		is compiled to run on a multi-processor computer (i.e., SGI).
	B or b if set to 0, suppresses the reporting of the pairwise
		alignments from the initial BLAST portion of the search.
	D or d if made non-zero, turns on the concise reporting of diagonals
		saved from the initial BLAST portion of the search.
	O or o if set to 0, suppresses the output that normally is sent to stderr
		for monitoring the program's progress.

 Thus a typical command is

	BLAST3 AABANK SEQ S=50 M=PAM250


Relevant literature:

	Karlin & Altschul, "Methods for assessing the statistical significance
	    of molecular sequence features by using general scoring schemes,"
	    Proc. Natl. Acad. Sci. USA 87 (1990) 2264-2268;
	Altschul and  Lipman, "Protein database searches for multiple alignments,"
	    Proc. Natl. Acad. Sci. USA 87 (1990) 5509-5513;
	Altschul, Gish, Miller, Myers and  Lipman, "Basic local alignment search
	    tool," J. Mol. Biol. 214 (1990) in press.

   (Note:  set tabstops every 4 columns for a better look).
*/
#include <ncbi.h>
#include <gish.h>
#include <gishlib.h>
#include <sys/stat.h>

#ifndef BLAST3
#define BLAST3
#endif
#define EXTERN
#include "blastapp.h"
#include "aabet.h" /* the amino acid alphabet */

#define DEFAULT_W	3	/* default word size */
#define DEFAULT_ehigh	0.1
#define DEFAULT_elow3	5.
#define DEFAULT_Efactor	2.
#define DEFAULT_High	0
#define DEFAULT_showdiags	0
#define DEFAULT_Low3	0
#define DEFAULT_M	"BLOSUM62"	/* default scoring matrix */
#define DEFAULT_T		11	/* default threshold for word hits */
#define DEFAULT_X		20	/* default constant for hit extensions */

char
	header[HEADERLEN_MAX],
	*qname = "NO NAME",
	*qtitle = "NO NAME",
	*dbtitle,
	*dbbuf,		/* Memory-resident database */
	*seqs, /* Large buffer in which to store all database sequences */
	*s_seq0, *s_seq, *ss_seq,
	*_q_seq,
	*q_seq,
	*qq_seq,
	M[FILENAME_MAX] = DEFAULT_M; /* name of file containing substitution scores */

int		qnamelen = 7;	/* length of the query's name */
int		qtitlelen = 7;
int		maxnamelen; /* length of longest name of sequence */

long	totdblen;	/* total no. of letters in the database */

unsigned long
	maxlen,		/* max length of a database sequence */
	numwords,	/* number of critical words in hash table */
	numexact;	/* number of exact words that did not score at least T */

Score_t
	S = DEFAULT_S,			/* cutoff for HSPs */
	S2;

int
	W = DEFAULT_W,			/* word size */
	W_set,
	S_set,
	E_set,
	T = DEFAULT_T,			/* threshold for word hits */
	T_set,
	X = DEFAULT_X,			/* constant for hit extensions */
	X_set,
	Y_set,
	Z_set,
	minusX = -DEFAULT_X;	/* -X */

double	*prob;
double	fq1[32];
struct record	firstrec[RECORD_MAX], *rec = firstrec;
char	**pts;
char	**names;
long	*lens;
double	E = DEFAULT_E,
		Lambda3, K3, H3,
		ehigh = DEFAULT_ehigh,
		elow3 = DEFAULT_elow3,
		Efactor = DEFAULT_Efactor;

int		showdiags = DEFAULT_showdiags,
		nrec;

Score_t
		High = DEFAULT_High,
		Low3 = DEFAULT_Low3,
		lows,
		highs;

Score_t	high_score, score_hist_dim;

/* data structure to hold locations of matches for current database sequence */

DIAG_TYPE
	*_dlev,			/* extent of discovered matches .. */
	*diag_level;	/*     on a given diagonal */

DIAG_PTR
	diag_stack,	/* list of diagonals with matches */
	diag_top;		/* next diagonal with a match */

DFAPtr	dp;	/* the automaton */

int		alphasize;			/* alpabet size */
int		minchar, maxchar;	/* min. and max. letter values */
static char
	/* array of amino acid names, plus null terminator */
	achars[ALPHASIZE_MAX+1],
	/* order[i][j] = j-th best char to sub. for i */
	order[ALPHAVAL_MAX+1][ALPHASIZE_MAX];

Score_t
	/* max cost of substituting for a character */
	maxcost[ALPHAVAL_MAX+1],
	/* matrx[i][j] = cost of replacing i by j */
	matrx[ALPHAVAL_MAX+1][ALPHAVAL_MAX+1];

Nlm_Boolean	consistency_flag = TRUE;

static void	initdb PROTO((char *dbname));
static void	search_aa PROTO((MPPROTO));
static int	get_seq PROTO((MPPROTOC unsigned long));
static void	dosearch PROTO((Nlm_VoidPtr,TaskBlkPtr));
static int	print_p PROTO((HitListPtr));

unsigned long	dfanstates, dfastatesize, dfasize, dfaextent;

Cputime_t	t0, t1, t2, t3, t4;

int		alphasize;				/* alpabet size */
int		minchar, maxchar;	/* min. and max. letter values */

Score_t
	maxcost[ALPHAVAL_MAX+1],	/* max cost of substituting for a character */
	/* matrx[i][j] = cost of replacing i by j */
	matrx[ALPHAVAL_MAX+1][ALPHAVAL_MAX+1];


/* UNIX-specific input routine; database must be pre-processed by SETDB */
static MFILE
	*sfile;		/* file pointer for sequences */

unsigned long	*seq_beg; 	/* starts of sequences */

char	*citation = "Altschul, Stephen F. and David J. Lipman (1990). Protein database searches for multiple alignments, Proc. Natl. Acad. Sci. USA 87:5509-5513.";

/*
Version	Date		Comment
-------	--------	---------------------------------------------------------
1.1.6   12-7-90         Using the improved DFA library code
1.1.7   12-18-90        Using new HitList structures; sharing more program code
1.1.8   1-14-91         Fixed bug in bldaa.c regarding perfect match words
1.1.9   1-16-91         Fixed bug in Poisson P-value calculation
1.1.10  1-28-91         Display one-line description of each db sequence hit
1.1.11  2-4-91          Word-wrap long sequence descriptions
1.1.12  3-13-91         Added calculation of default T in cases when W==3
1.1.13  5-1-91          ??
1.1.14  6-27-91         Fixed uninitialized variable maxdg in out3 and added
                        reporting of Lambda, K, and H in print_stats().
1.1.15	22-7-91	Default X is calculated using information theory
1.1.16	13-9-91	Changed behavior of parameters V and B.
1.1.17	23-9-91	Small change in extend_aa. (Use "S" instead of "MPPTR cutoff").
1.1.18	25-9-91	Improved reporting of individual HSP statistics
1.2.0	1-10-91	Improved Poisson statistics, screening and sorting of hits.
				(Actually there is no change to BLAST3 except the report
				format for HSPs.  Its version number is bumped mainly to
				keep version numbers for all programs more closely in sync).
1.2.1	11-11-91	pts[] array was not initialized to NULL values!
1.2.2	12-11-91	fixed bug in print_p(), which now uses the sequences
					that have already been read
1.2.3	23-12-91	Improved command line parsing, new -overlap option
1.2.4	6-1-92	Only use unambiguous (non-X) letters when calculating K & L
1.2.5	18-2-92	Switched to new dfa library
1.2.6	9-3-92	Faster K calculation in karlin()
1.2.7	18-3-92	Fixed bug in out3() when there are no HSPs to merge
1.2.8	31-3-92	Added a gap character '-' to the amino acid alphabet
1.2.9	12-6-92	Fixed bug in lambdak() when smaller alphabets are used
1.2.10	25-8-92	DEC Ultrix compatibility
1.2.11	4-9-92	Fixed bug in lib/hsppool.c
1.3.0	22-11-92	Added consistency of HSP Poisson Event counts
1.3.1   23-4-93 Added consistent Poisson P-value calculation
1.3.2   5-5-93  Karlin H was assumed to have wrong units in lib/stolen.c
1.3.3	9-6-93	Added gapdecayrate suggested by Phil Green
1.3.4	21-6-93	Expect values for HSPs now correspond to the Poisson P-values
1.3.5   22-6-93 Fixed bug in consistp.c implementation of R(i,3) -- Phil Green
1.3.6   25-6-93 Finished gapdecayrate -- modified cutoffs()
1.3.7	7-7-93	Added -overlap2 option, prompted by Erik Sonnhammer
1.3.8	29-10-93	Added -dbrecmin and -dbrecmax options
*/
CharPtr	version = "1.3.8";
CharPtr	rel_date = "29-Oct-93";

int
main(argc, argv)
	int		argc;
	char	**argv;
{
	char	*a, o = NULLB, *theArg;
	register int	i, j, k, ac, arglen;
	register char	*cp, ch;
	size_t	nheaders;
	unsigned long	qlen;

	sys_times(&t0);
	alarmprocset();

	banner(argv[0], version, rel_date, citation, (char *)NULL);

	if (argc < 3)
		usage();

	for (ac = 3; ac < argc; ++ac) {
		arglen = strlen(theArg = argv[ac]);
		if (str_casecmp("-over2", theArg) == 0 || str_casecmp("-overlap2", theArg) == 0) {
			ovlap = ovlap_a2;
			overlap_option = theArg;
			continue;
		}
		if (arglen >= 5 && str_ncasecmp(theArg, "-over", 5) == 0) {
			ovlap = ovlap_0;
			overlap_option = theArg;
			continue;
		}
		if (arglen >= 7 && str_ncasecmp("-filter", theArg, 7) == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing filter specification");
			filtercmd = pick_filter(argv[ac], QUERY_SEQTYPE);
			continue;
		}
		if (arglen >= 6 && arglen <= 11 &&
				str_ncasecmp("-echofilter", theArg, arglen) == 0) {
			echofilter_flag = TRUE;
			continue;
		}
		if (str_ncasecmp(theArg, "-consis", 7) == 0) {
			consistency_flag = FALSE;
			continue;
		}
		if (str_ncasecmp(theArg, "-poisson", 8) == 0) {
			continue;
		}
		if (str_casecmp(theArg, "-gi") == 0) {
			gi_option = TRUE;
			continue;
		}
		if (str_casecmp(theArg, "-dbrecmax") == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing -dbrecmax value");
			if (sscanf(argv[ac], "%ld", &dbrecmax) != 1 || dbrecmax <= 0)
				bfatal(ERR_INVAL, "Invalid -dbrecmax value:  %s", argv[ac]);
			continue;
		}
		if (str_casecmp(theArg, "-dbrecmin") == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing -dbrecmin value");
			if (sscanf(argv[ac], "%ld", &dbrecmin) != 1 || dbrecmin <= 0)
				bfatal(ERR_INVAL, "Invalid -dbrecmin value:  %s", argv[ac]);
			--dbrecmin;
			continue;
		}
		if (str_ncasecmp("-gapdecayrate", theArg, 8) == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing gapdecayrate");
			if (sscanf(argv[ac], "%lg", &gapdecayrate) != 1)
				bfatal(ERR_INVAL, "Invalid value specified with -gapdecayrate:  %s", argv[ac]);
			continue;
		}
		if (theArg[1] == '=')
			o = theArg[0];
		else
			if (theArg[0] == '-')
				o = theArg[1];
			else
				bfatal(ERR_INVAL, "Argument %d (``%s'') was unexpected or has improper form.", ac, theArg);
		a = theArg + 2;
		if (*a == NULLB) /* ignore */
			continue;
		if (islower(o))
			o = toupper(o);
		switch (o) {
			case 'W':
				if (sscanf(a, "%d", &W) != 1)
					bfatal(ERR_INVAL, "Bad value for W");
				if (W <= 0)
					bfatal(ERR_WORDSIZE, "Invalid value for W");
				if (W > WORDSIZE_MAX)
					bfatal(ERR_WORDSIZE, "W is too big--must be less than %d", WORDSIZE_MAX+1);
				W_set = TRUE;
				break;
			case 'P':
#ifdef MPROC_AVAIL
				numprocs = atoi(a);
				if (numprocs <= 0)
					bfatal(ERR_INVAL, "Bad value for P (the number of processors)");
				if (numprocs > THREADS_MAX)
					bfatal(ERR_INVAL, "Too many processors requested (Pmax = %d)", THREADS_MAX);
				numprocs = MIN(numprocs, THREADS_MAX);
#else
				printf("NOTE:  Parameter P is invalid--\n       this program has not been compiled to use multiprocessing\n");
#endif
				break;
			case 'S':
				if (sscanf(a, "%d", &S) != 1 || S < 1)
					bfatal(ERR_INVAL, "Bad value for S");
				S_set = TRUE;
				break;
			case 'M':
				if ((int)strlen(a) >= FILENAME_MAX)
					bfatal(ERR_INVAL, "File name is too long");
				strcpy(M,a);
				break;
			case 'T':
				if (sscanf(a, "%d", &T) != 1 || T < 1)
					bfatal(ERR_INVAL, "Bad value for T");
				T_set = TRUE;
				break;
			case 'X':
				if (sscanf(a, "%d", &X) != 1 || X < 1)
					bfatal(ERR_INVAL, "Bad value for X");
				minusX = -X;
				X_set = TRUE;
				break;
			case 'E':
				if (sscanf(a, "%lg", &E) != 1 || E < 1.)
					bfatal(ERR_INVAL, "E must be at least 1.");
				E_set = TRUE;
				if (E < 200.)
					warning("The chosen value for E is unusually low.  The default value is %lg.\n", DEFAULT_E);
		    	break;
			case 'U':
				High=atoi(a);
		    	break;
			case 'R':
				if (sscanf(a, "%lg", &Efactor) != 1 || Efactor <= 0.)
					bfatal(ERR_INVAL, "Bad value for R");
				break;
			case 'N':
				if (sscanf(a, "%lg", &ehigh) != 1 || ehigh <= 0.)
					bfatal(ERR_INVAL, "Bad value for N");
		    	break;
			case 'L':
				if (sscanf(a, "%d", &Low3) != 1 || Low3 < 1)
					bfatal(ERR_INVAL, "Bad value for L");
		    	break;
			case 'F':
				if (sscanf(a, "%lg", &elow3) != 1 || elow3 <= 0.)
					bfatal(ERR_INVAL, "Bad value for F");
		    	break;
			case 'Y':
				Y_set = TRUE;
				if (sscanf(a, "%d", &Meff) != 1 || Meff == 0)
					bfatal(ERR_INVAL, "Bad value for Y");
		    	break;
			case 'Z':
				Z_set = TRUE;
				if (sscanf(a, "%d", &Neff) != 1 || Neff == 0)
					bfatal(ERR_INVAL, "Bad value for Z");
		    	break;
			case 'V':
				if (sscanf(a, "%d", &V) != 1 || V < 0)
					bfatal(ERR_INVAL, "Bad value for V");
				V_set = TRUE;
				break;
			case 'B':
				if (sscanf(a, "%ld", &showblast) != 1 || showblast < 0)
					bfatal(ERR_INVAL, "Bad value for B");
				break;
			case 'H':
				showhist = atoi(a);
				break;
			case 'D':
				showdiags = atoi(a);
				break;
			default:
				printf("Bad option:  %c", o);
				usage();
		}
	}

	/* read file of substitution weights */
	get_weights(M, AAID_MAX+1, aa_atob, &alphasize, &minchar, &maxchar, &achars,
					&order, &lows, &highs, &maxcost, &matrx);

	get_query(argv[2]);
	(void) wrap(stdout, "Query=  ", qtitle, qtitlelen, 79, qnamelen+1);
	(void) printf("        (%s letters)\n", lu2s(q_len));

	if (seqfilter(filtercmd, aa_btoa, aa_atob,
				qtitle, qtitlelen,
				(UcharPtr)q_seq, q_len,
				(UcharPtr)q_seq, (size_t PNTR)&q_len) != 0)
		bfatal(ERR_UNDEF, "Non-zero return code from sequence filter");

	putchar('\n');
	initdb(argv[1]); /* Open database files */
	(void) fflush(stdout);

	pts = (char **)ckalloc0(nbr_seqs*sizeof(*pts));
	names = (char **)ckalloc(nbr_seqs*sizeof(*names));
	lens = (long *)ckalloc(nbr_seqs*sizeof(*lens));

	prob = (double *)ckalloc((3*highs-3*lows+1) * sizeof(*prob));
	for (i=0; i<DIM(fq1); ++i)
		fq1[i] = 0.0;
	for (i=0, qlen = q_len; i<q_len; ++i) {
		if (q_seq[i] != aa_atob[UNKNOWN_AA_CHR])
			++fq1[q_seq[i]];
		else
			--qlen;
	}
	if (!Y_set)
		Meff = qlen;
	for (i=0; i<DIM(fq1); ++i)
		fq1[i] /= qlen;
	for (i=highs-lows; i>=0; --i)
		prob[i] = 0.0;
	for (i=minchar; i<=maxchar; ++i)
		for (j=1;j<DIM(aafq);++j) {
			if (matrx[i][j] > SCORE_MIN/WORDSIZE_MAX)
				prob[matrx[i][j]-lows] += fq1[i]*aafq[j];
		}
	karlin(lows, highs, prob, !L_set, &Lambda, !K_set, &K, TRUE, &H);

	if (!X_set) {
		X = ceil(10./(Lambda/LN2));
		minusX = -X;
	}

	if (S_set && !E_set)
		E = 0.;
	cutoffs(&S, &E, &highs_exp, Lambda, K, H, Neff, Meff, gapdecayrate);

	score_hist_dim = etos(0.01, Lambda, K, H, Neff, Meff) + 1;
	score_hist = (unsigned long *)ckalloc0(sizeof(score_hist[0])*(score_hist_dim+1));
	poolp = HSPPoolNew();
	hl_pool = HitListPoolNew();

	if (!T_set) {
		T = DEFAULT_T;
		/* Ad hoc equations for PAM matrices with scale = ln(2)/2 */
		if (W == 3)
			T = ceil((4.376 + 1.3*log(H))/Lambda - 0.5);
			/*T = ceil((2.0 + 2.741*H)/Lambda - 0.5);*/
		if (W == 4)
			T = ceil((2.1 + 3.751*H)/Lambda - 0.5);
		/*
		The following adjustment to T is tightly linked
		to the PAM120 substitution matrix.
		*/
		if (S < 35)
			T -= 1;
	}

	S2 = S;

	lows *= 3;
	highs *= 3;
	for (i=highs-lows;i>=0;--i)
		prob[i] = 0.0;
	for (i=minchar;i<=maxchar;++i)
		for (j=1;j<DIM(aafq);++j)
			for (k=1;k<DIM(aafq);++k) {
				if (matrx[i][j] > SCORE_MIN/WORDSIZE_MAX)
					prob[matrx[i][j]+matrx[i][k]+matrx[j][k]-lows] +=
						fq1[i]*aafq[j]*aafq[k];
			}
	karlin(lows, highs, prob, !L_set, &Lambda3, !K_set, &K3, TRUE, &H3);
	mem_free((char *)prob);

	sys_times(&t1);
	dp = bld_table_aa((DFA *)NULL, INT_MIN, q_seq, q_len, W, T, alphasize,
						&matrx, &maxcost, &order, &numwords, &numexact);
	end_table(dp);
	sys_times(&t2);

	maxnamelen = qnamelen;

	RunWild("Searching", dbreccnt, dosearch);

#ifdef SIGXCPU
	/* Report results without interruption */
	signal(SIGXCPU, SIG_IGN);
#endif

	nrec = rec - firstrec;

	/* Retrieve header lines for matching sequences */
	for (i=0; i<nbr_seqs; ++i) {
		if (pts[i] == 0)
			continue;
		/* get the header line */
		if (mfil_seek(hfile, header_beg[i], SEEK_SET) == -1)
			bfatal(ERR_DBASE, "fseek error on header file");
		if (mfil_read(header, 1, sizeof(header)-1, hfile) <= 0)
			bfatal(ERR_DBASE, "error reading header");
		header[sizeof(header)-1] = NULLB;
		cp = header;
		while ((ch = *cp) != NULLB && !isspace(ch))
			++cp;
		j = cp - header;
		if (j > maxnamelen)
			maxnamelen = j;
		names[i] = cp = (char *)ckalloc(j+1);
		cp[j] = NULLB;
		while (j-- > 0)
			cp[j] = header[j];
	}

	dfanstates = dp->nstates;
	dfastatesize = dp->statesize;
	dfasize = dfa_size(dp);
	dfaextent = dfa_extent(dp);
	dfa_destruct(dp); /* Deallocate the DFA */

	collate_hits();

	hasHSP = hitlist_count(firsthl);
	sort_hitlists(&firsthl, cmp_by_pval);

	if (showblast || V) {
		HitListPtr	hlp;
		size_t	cnt;
		long	hashsp;
		double	phigh;

		nheaders = MIN(hasHSP, MAX(V, showblast));
		phigh = etop(ehigh);
		for (cnt=0, hlp = firsthl;
				hlp != NULL && hlp->signifhsp->P_val <= phigh;
				hlp = hlp->next)
			++cnt;

		nheaders = MIN(cnt, nheaders);
		if (nheaders < hasHSP)
			hitlist_truncate(&firsthl, nheaders);
		get_headers(DATABASE_SEQTYPE, firsthl);

		print_hist(score_hist_dim, score_hist, S, T, Lambda, K, H,
						Meff, Neff, nbr_seqs);

		hashsp = hasHSP; /* save */
		hasHSP = cnt;
		print_headers(firsthl, phigh, FALSE, 80);

		print_HSPs(firsthl, phigh, print_p);
		hasHSP = hashsp; /* restore */
	}
	putchar('\n');

	sys_times(&t3);
	out3();

	alarm(0);
	sys_times(&t4);
	print_parms();
	print_stats();
	ckwarnings();
	exit_code(ERR_NONE);
	/*NOTREACHED*/
}

print_parms()
{
	(void) printf("\n\nParameters:\n");
	if (filtercmd != NULL && filtercmd[0] != NULLB)
		(void) printf("  Query filter:  \"%s\"\n", filtercmd);
    (void) printf("  E = %#0.2lg, S = %ld (%.1lf bits),  E2 = n/a, S2 = n/a\n",
        E, (long)S, (double)S*Lambda/LN2
        );
    (void) printf("  W = %d, T = %ld (%.1lf bits), X = %ld (%.1lf bits)\n",
        W, (long)T, (double)T*Lambda/LN2,
        (long)X, (double)X*Lambda/LN2);
    (void) printf("  M = %s\n", M);
	(void) printf("  H = %d, V = %d, B = %d\n", showhist, V, showblast);
	report_gapdecayrate();
	if (overlap_option != NULL)
		printf("  Overlap option:  %s\n", overlap_option);
	return ferror(stdout);
}

print_stats()
{
	printf("\nStatistics:\n");
	printf("  Lambda = %#0.3lg, Lambda/ln2 = %#0.3lg, K = %#0.3lg, H = %#0.3lg bits/position\n",
				Lambda, Lambda/LN2, K, H/LN2);

    (void) printf("  Expected/Observed high score = %s (%.1lf bits) / %s (%.1lf bits)\n",
            ld2s(highs_exp), ((double)highs_exp)*(Lambda/LN2),
            ld2s(highs_obs), ((double)highs_obs)*(Lambda/LN2)
            );

	printf("  # of letters in query = %s", lu2s(q_len));
	if (Meff != q_len)
		printf("  (Y = %s)\n", lu2s(Meff));
	else
		putchar('\n');
	printf("  # of neighborhood words:  %s\n", ld2s(numwords));
	(void) wrap(stdout, "  Database:  ", dbtitle, -1, 79, 13);
	printf("  # of letters in db:  %s", ld2s(totdblen));
	if (Neff != totdblen)
		printf("  (Z = %s)\n", lu2s(Neff));
	else
		putchar('\n');
	printf("  # of word hits against database:  %s\n", ld2s(numhits));
	printf("  # of failed hit extensions:  %s\n", ld2s(numfailed));
	printf("  # of successful extensions:  %s\n", ld2s(totHSP+overcnt));
	printf("  # of overlapping HSPs discarded:  %s\n", ld2s(overcnt));
	printf("  # of HSPs reportable from BLAST phase:  %s\n", ld2s(totHSP));
	printf("  # of sequences in database:  %s\n", ld2s(nbr_seqs));
	printf("  # of database sequences satisfying E:  %s\n",
		ld2s(hasHSP));

	printf("No. of states in DFA:  %s (%s KB)\n",
			lu2s(dfanstates),
			lu2s(HOWMANY(dfastatesize*dfanstates, KBYTE)));
	printf("Total size of DFA:  %s KB (%s KB)\n",
			lu2s(HOWMANY(dfasize, KBYTE)),
			lu2s(HOWMANY(dfaextent, KBYTE)));
	sys_showtime("Time to generate neighborhood", &t1, &t2, stdout);
#ifdef MPROC_AVAIL
	printf("No. of processors used:  %d\n", numprocs);
#endif
	sys_showtime("Time to search database", &t2, &t3, stdout);
	sys_showtime("Time to find 3-alignments", &t3, &t4, stdout);
	sys_showtime("Total cpu time", (Cputime_t *)NULL, &t4, stdout);
}

static void
dosearch(userp, tp)
	Nlm_VoidPtr	userp;
	TaskBlkPtr	tp;
{
	int		j, k;
	MPDEF

#ifdef MPROC_AVAIL
	initmpc(tp->id, mp = &mpc[tp->id]);
#endif
	/* For each sequence in the database */
	while (TaskNext(tp) >= 0) {
		get_seq(MPC tp->task_cur + dbrecmin);
		MPPTR high_score = 0;
		search_aa(MP);	/* search a database sequence */

		if (MPPTR high_score <= score_hist_dim)
			++MPPTR score_hist[MPPTR high_score];
		else
			++MPPTR score_hist[score_hist_dim];

		if (MPPTR curhl.hp != NULL) {
			register HSPPtr	hp = MPPTR curhl.hp;

			k = INT_MIN;
			M_LOCK;
			do {
				j = hp->s_pos - hp->q_pos;
				if (j == k)	/* Remember only one HSP from a given diagonal */
					continue;
				if (rec >= &firstrec[DIM(firstrec)])
					bfatal(ERR_RECORD3, "Out of storage for diagonals; decrease E parameter, or increase RECORD_MAX and recompile");
				rec->dg2 = k = j;
				rec->score = hp->score;
				rec->start1 = hp->q_pos;
				rec->length = hp->len;
				rec->seq2 = tp->task_cur + dbrecmin;
				rec->seq3 = -1;
				++rec;
			} while ((hp = hp->next) != NULL);
			M_UNLOCK;
			pts[tp->task_cur + dbrecmin] = MPPTR s_seq;
			lens[tp->task_cur + dbrecmin] = MPPTR s_len;
			sort_HSPs(&MPPTR curhl);
			if (consistency_flag) {
				consistn(&MPPTR curhl);
				consist_pvals(&MPPTR curhl, Meff, MPPTR s_len, Neff);
			}
			else {
				pcnt(&MPPTR curhl);
				pvals(&MPPTR curhl, Meff, MPPTR s_len, Neff);
			}
			MPPTR curhl.seq_nbr = tp->task_cur + dbrecmin;
			hitlist_save(MPPTR hl_pool, &MPPTR firsthl, &MPPTR curhl);
			MPPTR totHSP += MPPTR curhl.hspcnt;
			MPPTR curhl.hp = NULL;
			MPPTR curhl.hspcnt = 0;
		}
	}
}

#ifdef MPROC_AVAIL
/* Initialize the processor context pointed to by mp */
initmpc(id, mp)
	int		id;
	MPContextPtr	mp;
{
	int		len;
	register int	i;

	mp->id = id;

	len = maxlen + q_len + 1;
	mp->_dlev = (DIAG_TYPE *)ckalloc0(sizeof(*_dlev)*len);
	mp->diag_level = mp->_dlev + q_len;
	mp->diag_top = mp->diag_stack = (int *)ckalloc(sizeof(*diag_top)*len);
	mp->s_seq0 = ckalloc(maxlen+2);
	*mp->s_seq0++ = NULLB;
	mp->s_seq = mp->s_seq0;
	if (id == 0) {
		mp->score_hist = score_hist;
		mp->poolp = poolp;
		mp->hl_pool = hl_pool;
	}
	else {
		mp->score_hist = (unsigned long *)ckalloc0(sizeof(mp->score_hist[0])*(score_hist_dim+1));
		mp->poolp = HSPPoolNew();
		mp->hl_pool = HitListPoolNew();
	}
}
#endif /* MPROC_AVAIL */

/* get_query - read the query sequence */
get_query(filename)
	char *filename;
{
	char	*base, *bp;
	size_t	basemax;
	FILE	*fp;
	register char *p, ch;
		
	fp = ckopen(filename, "r", 1);

	base = ckalloc(basemax = 10*KBYTE);
	bp = base + 1;

	while (vfgets(&base, &basemax, &bp, fp) != NULL) {
		if (*bp == '>') {	/* Optional header/title line */
			if (bp != base+1)/* Second sequence encountered */
				break;
			p = str_chr(bp, '\n');
			if (p != NULL)
				*p = NULLB;
			qtitle = str_dup(bp+1);
			qtitlelen = strlen(qtitle);
			for (p=qtitle; (ch = *p) && !isspace(ch); ++p)
				;
			*p = NULLB;
			qname = str_dup(qtitle);
			*p = ch;
			qnamelen = strlen(qname);
			continue;
		}
		p = bp;
		while ((ch = *p++) != NULLB) {
			if (aa_atob[ch] <= AAID_MAX) {
				if (islower(ch))
					ch = toupper(ch);
				*bp++ = ch;
			}
			else
				if (aa_atob[ch] == AAID_NAR) {
					aa_atob[ch] = AAID_IGNORE;
warning("Invalid amino acid code ``%c'' encountered in query sequence will be skipped.", ch);
				}
		}
		if (bp - base > QUERYLEN_MAX-1)
			bfatal(ERR_QUERYLEN, "query sequence exceeds %d characters in length",
				QUERYLEN_MAX-1);
			
	}
	(void) fclose(fp);
	_q_seq = base;
	_q_seq[0] = NULLB;
	q_seq = base+1;
	q_len = bp - q_seq;
	if (q_len < W)
		bfatal(ERR_QUERYLEN, "query sequence is shorter than the word length, W=%d.", W);
	*bp = NULLB;

	(void) ckseqtype(QUERY_SEQTYPE, q_seq, q_len);

	p = q_seq;
	while ((ch = *p) != NULLB)
		*p++ = aa_atob[ch];

	qq_seq = ckalloc(q_len+1);
	ss_seq = ckalloc(q_len+1);
	return 0;
}

/* UNIX-specific input routine; database must be pre-processed by SETDB */

static void
initdb(dbname)
	char	*dbname;
{
	struct stat	sbuf;
	MFILE	*tfile;
	long	dbtype, dbformat, i;
	char	*dirname = "";
	char	name[FILENAME_MAX];
	int		many, len;
	int		once=0;
	off_t	dbsize;

RETRY:
	if (*dirname != NULLB)
		(void) sprintf(dbfname, "%s%s%s", dirname, DIRDELIMSTR, dbname);
	else
		SAFENCPY(dbfname, dbname, sizeof(dbfname));
	(void) sprintf(name, "%s%s", dbfname, HEADER_EXT);
	hfile = mfil_open(name, "r", MFIL_OPT_ALL);
	if (hfile == NULL) {
		if ((dirname = getenv("BLASTDB")) == NULL)
			dirname = BLASTDB;
		if (++once == 1)
			goto RETRY;
		bfatal(ERR_FOPEN, "Couldn't open database file ``%s''.  Make sure you have specified the correct database name and requested a database of the correct type (protein) for use with this program.", basename(name, NULL));
	}

	(void) sprintf(name, "%s%s", dbfname, TABLE_EXT);
	tfile = mfil_open(name, "r", MFIL_OPT_ALL);
	if (tfile == NULL)
		bfatal(ERR_FOPEN, "Couldn't open database file ``%s''.", name);
	if (mfil_dup_long(tfile, &dbtype, 1) != 1)
		bfatal(ERR_DBASE, "Database read error on file", name);
	if (dbtype != DB_TYPE_PRO)
		bfatal(ERR_DBASE, "``%s'' is not a protein sequence database", dbname);
	if (mfil_dup_long(tfile, &dbformat, 1) != 1)
		bfatal(ERR_DBASE, "Database read error on file", name);
	if (dbformat != AAFORMAT)
		bfatal(ERR_DBASE, "Database format is an incompatible version; re-run ``setdb''.");
 
	(void) mfil_dup_long(tfile, &i, 1);
	dbtitle = mfil_get(tfile, i);
	if (i%4 != 0) {                                      
		i = 4 - (i%4);
		mfil_seek(tfile, i, SEEK_CUR);
	}                           

	(void) mfil_dup_long(tfile, (long *)&nbr_seqs, 1);
	(void) mfil_dup_long(tfile, (long *)&totdblen, 1);
	if (mfil_dup_long(tfile, (long *)&maxlen, 1) != 1)
			bfatal(ERR_DBASE, "corrupted database file:  %s", name);

	if (dbtitle[0] == NULLB)
		dbtitle = str_dup(dbname);
	(void) wrap(stdout, "Database:  ", dbtitle, -1, 79, 11);
	(void) printf("           %s sequences; %s total letters.\n",
			ld2s(nbr_seqs), ld2s(totdblen));                       

	if (nbr_seqs <= 0 || totdblen <= 0)
		bfatal(ERR_DBASE, "Nothing in the database to search!?");
 
	ss_seq = (char *)ckalloc(sizeof(*ss_seq)*(maxlen+1));

	many = nbr_seqs + 1;
	seq_beg = (unsigned long *)mfil_get_long(tfile, many);
	header_beg = (unsigned long *)mfil_get_long(tfile, many);
	if (seq_beg == NULL || header_beg == NULL)
		bfatal(ERR_DBASE, "improper offsets in database table file");

	(void) sprintf(name, "%s%s", dbfname, AA_SEARCHSEQ_EXT);
	/* Attach a shared memory segment for the .bsq file (if one exists) */
	sfile = mfil_open(name, "r", MFIL_OPT_ALL);
	if (sfile == NULL)
		bfatal(ERR_FOPEN, "Couldn't open database file %s", name);
	if ((s_seq0 = dbbuf = (CharPtr)MFILE_SP(sfile)) == NULL) {
		fflush(stdout);
		fstat(fileno(MFILE_FP(sfile)), &sbuf);
		dbsize = sbuf.st_size;
		s_seq0 = (char *)ckalloc(sizeof(*s_seq)*dbsize);
		if (mfil_read((char *)s_seq0, (int)dbsize, 1, sfile) != 1)
			bfatal(ERR_DBASE, "Database read error");
		(void) mfil_close(sfile);
	}


	len = maxlen + q_len + 1;
	_dlev = (DIAG_TYPE *)ckalloc0(sizeof(*_dlev)*len);
	diag_level = _dlev + q_len;
	diag_top = diag_stack = (int *)ckalloc(sizeof(*diag_top)*len);

	if (Neff == 0)
		Neff = totdblen;

	dbrecmin = MAX(dbrecmin, 0);
	dbrecmin = MIN(dbrecmin, nbr_seqs);
	dbrecmax = (dbrecmax > 0 ? MIN(dbrecmax, nbr_seqs) : nbr_seqs);
	dbreccnt = dbrecmax - dbrecmin;
	if (dbreccnt < nbr_seqs)
		printf("           Subset of database from %ld to %ld.\n",
			dbrecmin+1, dbrecmax);
}

/* get_seq - read a database sequence and header line, return 0 on EOF */
static int
get_seq(MPC current)
	MPDEF
	unsigned long	current;
{
	if (dbbuf != NULL) {
		MPPTR s_len = seq_beg[current+1] - seq_beg[current] - 1;
		MPPTR s_seq = dbbuf + seq_beg[current];
	}
	else {
		MPPTR s_seq = s_seq0 + seq_beg[current];
		MPPTR s_len = seq_beg[current+1] - seq_beg[current] - 1;
	}
	if (MPPTR s_len > maxlen)
		bfatal(ERR_DBASE, "Database sequence longer than expected maxlen!");
	return 0;
}

/*****************************/
#include "searcha.inc"
/*****************************/

static int
print_p(hlp)
	HitListPtr	hlp;
{
	Seqid_t	current;
	Coord_t	len;
	int		i;
	HSPPtr	hp;
	register char	*cp, ch;

	current = hlp->seq_nbr;

	cp = hlp->header;
	while ((ch = *cp) != NULLB && !isspace(ch))
		++cp;
	while ((ch = *cp) != NULLB && isspace(ch))
		++cp;
	i = cp - hlp->header + 1;
	i = MIN(i, 12);
	putchar('\n');
	putchar('\n');
	(void) wrap(stdout, ">", hlp->header, hlp->header_len, 79, i);

	/* get sequence */
	/* Unlike blastp, the entire database is always already in memory */
	s_seq = s_seq0 + seq_beg[current];
	len = seq_beg[current+1] - seq_beg[current] - 1;

	(void) printf("%*sLength = %s\n", i, "", ld2s((long)len));
	
	for (hp = hlp->hp; hp != NULL; hp = hp->next) {
		if (hsp_print(stdout, hp, "", (UcharPtr)q_seq + hp->q_pos, (UcharPtr)s_seq + hp->s_pos, matrx, aa_btoa, aa_btoa, q_len, s_len, nox_coord, nox_coord, nox_coord, nox_coord, NULL, NULL, Lambda) != 0)
			return -1;
	}
	return 0;
}

usage()
{
	fflush(stdout);
	fflush(stderr);
	fprintf(stderr, "\nUsage:\n\n");
	fprintf(stderr,
	"    %s database query_sequence [options]\n\n", module);
	fprintf(stderr,
		"Valid %s options:  E, S, W, T, X, M, Y, Z, F, L, R, N, U, H, V, B, D and O\n",
		module);
	fprintf(stderr, "\t-dbrecmin #\n");
	fprintf(stderr, "\t-dbrecmax #\n");
	fprintf(stderr, "\t-gi (gi identifier suppression)\n");
	fprintf(stderr, "\t-overlap\n\t-overlap2\n\t-gapdecayrate #\n\t-filter filteroption\n\t-echofilter\n\t-consistency\n");
	fprintf(stderr, "\t-sort_by_pvalue\n\t-sort_by_count\n\t-sort_by_highscore\n\t-sort_by_totalscore\n");
	exit_code(ERR_INVAL);
}
