/**************************************************************************
*                                                                         *
*                             COPYRIGHT NOTICE                            *
*                                                                         *
* This software/database is categorized as "United States Government      *
* Work" under the terms of the United States Copyright Act.  It was       *
* produced as part of the author's official duties as a Government        *
* employee and thus can not be copyrighted.  This software/database is    *
* freely available to the public for use without a copyright notice.      *
* Restrictions can not be placed on its present or future use.            *
*                                                                         *
* Although all reasonable efforts have been taken to ensure the accuracy  *
* and reliability of the software and data, the National Library of       *
* Medicine (NLM) and the U.S. Government do not and can not warrant the   *
* performance or results that may be obtained by using this software,     *
* data, or derivative works thereof.  The NLM and the U.S. Government     *
* disclaim any and all warranties, expressed or implied, as to the        *
* performance, merchantability or fitness for any particular purpose or   *
* use.                                                                    *
*                                                                         *
* In any work or product derived from this material, proper attribution   *
* of the author(s) as the source of the software or data would be         *
* appreciated.                                                            *
*                                                                         *
**************************************************************************/
/*
BLASTP - Search a protein database for regions that approximately match
regions of a query sequence.  The basic command syntax is

 	BLASTP database query_sequence

where:
 	"database" names a file containing protein sequences, separated by
	header lines beginning with the character '>'.  A typical
	sequence entry is:

		>CCHU (PIR) Cytochrome c - Human | 1.0,1.0,1.0,1.0,1.0
		GDVEKGKKIFIMKCSQCHTVEKGGKHKTGP
		NLHGLFGRKTGQAPGYSYTAANKNKGIIWG
		EDTLMEYLENPKKYIPGTKMIFVGIKKKEE
		RADLIAYLKKATNE

	The database must be pre-processed with the SETDB program.

	"query_sequence" names a file containing a protein sequence (where the
	header line is optional).

The BLASTP command permits optional additional arguments, such as X=50,
that reset certain internal parameters.  The available parameters are:

	E or e is the _expected_ no. of HSPs to produce (which influences S)
		(default is 25).
	S or s gives the High-scoring Segment Pair (HSP) cutoff score
		(default is calculated from E).
	T or t gives the threshold for a word match (default is calculated,
		but is usually 11 or 12 with the PAM120 matrix and W=3).
	X or x gives the value for terminating word extensions (default is 20)
	M or m is the name of a file containing substitution costs matrix
		(default is "BLOSUM62")
	Y or y is the effective length of the querY sequence for statistical
		significance calculations (default uses the actual query length)
	Z or z is the effective databaZe length for statistical significance
		calculations (default uses the actual database length)
	P or p is the number of processors to utilize when the program
		is compiled to run on a multi-processor computer (i.e., SGI).

Thus a typical command is

	blastp aabank myseq e=0.5 x=25

Literature:

    Karlin & Altschul, "Methods for assessing the statistical significance
        of molecular sequence features by using general scoring schemes,"
        Proc. Natl. Acad. Sci. USA 87 (1990) 2264-2268;
    Altschul, Gish, Miller, Myers and Lipman, "Basic local alignment search
        tool," J. Mol. Biol. 214 (1990) in press.

 (Note:  set tabstops every 4 columns for a better appearance).
*/
#include <ncbi.h>
#include <gish.h>
#include <gishlib.h>
#include <sys/stat.h>

#define EXTERN
#ifndef BLASTP
#define BLASTP
#endif
#include "blastapp.h"
#include "aabet.h"	/* the alphabet */

#define DEFAULT_M	"BLOSUM62"	/* default scoring matrix */
#define DEFAULT_T	13	/* default threshold for neighborhood word hits */
#define DEFAULT_E2	0.15	/* secondary Expect, when at least 1 HSP is found */
#define DEFAULT_S2	0	/* secondary cutoff, when at least 1 HSP is found */
#define DEFAULT_X	20	/* default score drop-off during hit extensions */
#define DEFAULT_W	3	/* default word size */

char
	*dbtitle,
	*qname = "NO NAME",
	*qtitle = "NO NAME",
	*dbbuf,			/* Memory-resident database */
	*s_seq0, *s_seq, *ss_seq,
	*_q_seq,	/* null prefix + query sequence + null terminator */
	*q_seq,		/* query sequence + null terminator */
	*qq_seq,	/* binary-encoded query sequence + null terminator */
	M[FILENAME_MAX] = DEFAULT_M;	/* file containing the substitution scores */

unsigned long
		totdblen;	/* total no. of letters in the database */

int		qnamelen = 7, qtitlelen = 7;

unsigned long
	maxlen,		/* max length of a database sequence */
	numwords,	/* number of critical words in hash table */
	numexact;	/* number of exact words that did not score at least T */

Score_t
	highs_dummy;


Score_t
	Ssave,
	S = DEFAULT_S,		/* cutoff score for High-scoring Segment Pairs (HSP) */
	S2 = DEFAULT_S2;	/* secondary cutoff score after at least 1 HSP found */

int
	S_set,
	S2_set,
	T = DEFAULT_T,		/* threshold for word hits */
	T_set,
	X = DEFAULT_X,		/* constant for word-hit extensions */
	minusX = -DEFAULT_X, /* -X */
	X_set,
	Y_set,
	W = DEFAULT_W,		/* word size */
	W_set,
	W2 = DEFAULT_W,
	W2_set,
	T2 = DEFAULT_T,
	T2_set,
	X2 = DEFAULT_X,
	X2_set,
	E_set,
	E2_set,
	M_set;

double
	E = DEFAULT_E,	/* Expected number of HSPs achieving cutoff score */
	E2 = DEFAULT_E2;

/* data structure to hold locations of matches for current database sequence */
DIAG_TYPE
	*_dlev,			/* flags for which diagonals... */
	*diag_level;	/* ...have been examined */

DIAG_PTR
	diag_stack,	/* list of diagonals with matches */
	diag_top;	/* next diagonal with a match */

static Score_t
	maxscore;		/* maximum achievable score */

static MFILE
	*sfile;			/* file pointer for sequences */

static Score_t
	highs,		/* highest score in the PAM matrix */
	lows;		/* lowest score in the PAM matrix */

unsigned long
	*seq_beg, 		/* starts of sequences in file or memory-resident dbase */
	*header_beg;	/* starts of headers */

DFAPtr	dp;	/* the automaton */

int		alphasize;			/* alpabet size */
int		minchar, maxchar;	/* min. and max. letter values */
static char
	/* array of amino acid names, plus null terminator */
	achars[ALPHASIZE_MAX+1],
	/* order[i][j] = j-th best char to sub. for i */
	order[ALPHAVAL_MAX+1][ALPHASIZE_MAX];

Score_t
	/* max cost of substituting for a character */
	maxcost[ALPHAVAL_MAX+1],
	/* matrx[i][j] = cost of replacing i by j */
	matrx[ALPHAVAL_MAX+1][ALPHAVAL_MAX+1];

Nlm_Boolean	consistency_flag = TRUE;

static void	initdb PROTO((char *dbname));
static void	dosearch PROTO((Nlm_VoidPtr userp, TaskBlkPtr tp));
static int	get_seq PROTO((MPPROTOC unsigned long));
static void	search_aa PROTO((MPPROTO));
static int	print_p PROTO((HitListPtr));

unsigned long	dfanstates, dfastatesize, dfasize, dfaextent;

Cputime_t	t0, t1, t2, t3;

char	*citation = "Altschul, Stephen F., Warren Gish, Webb Miller, Eugene W. Myers, and David J. Lipman (1990).  Basic local alignment search tool.  J. Mol. Biol. 215:403-410.";

/*
Version   Date          Modification
------- -------- ----------------------------------------------------
1.1.0            Added REFINED_STATS code to modify Expect.
1.1.1            Added REFINED_STATS code to modify Poisson P-values.
1.1.2            Modified the *.tab file to include the dbtype.
1.1.3            Renamed *.tab file to *.atb, and *.hed to *.ahd.
1.1.4  11-21-90  Corrected the calculation of Poisson P-values.
1.1.5  11-24-90  When at least one HSP is found involving a given database
                 sequence, it is scanned again with a lower cutoff score.
1.1.6  12-4-90   Fixed hit extension algorithm to halt when score <= 0
1.1.7  12-5-90   Added histogram of highest-scoring hit extensions
1.1.8  12-6-90   Moved many statistical calculations into library routines
1.1.9  12-7-90   Using the improved DFA library code
1.1.10 12-11-90  Now using linked lists of HSPs; better sorting of HSPs
1.1.11  1-14-91  Fixed bug in bldaa.c regarding perfect matches
1.1.12  1-15-91  Fixed bug in Poisson P-value calculations
1.1.13  1-16-91  Fixed another bug in Poisson P-value calculations
1.1.14  1-28-91  Display one-line description of each db sequence hit
1.1.15  2-4-91   Word-wrap sequence descriptions
1.1.16  2-25-91  Added the "Searching..." progress indicator
1.1.17  3-13-91  Added code to calculate default T when W==3.
1.1.18  5-5-91   Created parameter V; support for BLASTDB and BLASTPAM
                 environment variables; use of standard NCBI headers.
1.1.19   7-22-91   Default X is calculated using information theory
1.1.20   9-13-91   Changed behavior of parameters V and B.
1.1.21   9-23-91   Improved efficiency in search_aa() and extend_aa().
1.1.22   9-25-91   Improved reporting of individual HSP statistics
1.2.0   10-1-91   Improved Poisson statistics, sorting of hits
                 (Actually, there is no change to BLASTP other than the
                 report format, but its version is bumped to keep version
                 numbers more closely in sync between all of the programs.
1.2.1   26-12-91  Improved error reporting; new K & L options
1.2.2   6-1-92    Only use unambiguous letters (non-X) when calculating K & L
1.2.3   18-2-92   Switched to new dfa library
1.2.4   9-3-92    Faster K calculation in karlin()
1.2.5   31-3-92   Added gap character '-' to the amino acid alphabet
1.2.6   12-6-92   Fixed bug in lambdak() when smaller alphabets are used
1.2.7   16-6-92   Added several hitlist sorting options
1.2.8   25-8-92   DEC Ultrix compatibility
1.2.9   4-9-92    Fixed bug in lib/hsppool.c
1.2.10  26-10-92  Fixed bug in segmented sequence handling
1.3.0   16-11-92  Hitlists are pruned at the point where E is not satisfied;
                  S2 is the cutoff for manipulating HSPs, E and S determine
                  where to cutoff hit lists.
1.3.1   18-11-92  Added -filter option for query sequence filtering
1.3.2   20-11-92  Consistency feature, yields more realistic Poisson N count
1.3.3   23-4-93   Consistent Poisson P-values to go along with N count
1.3.4   5-5-93  Karlin H was assumed to have wrong units in lib/stolen.c
1.3.5   6-5-93  Tweaked DEFAULT_E2 upward to compensate for yesterday's bugfix
1.3.6	9-6-93	Added gapdecayrate suggested by Phil Green
1.3.7	19-6-93 Expect values correspond now to the P-values reported
1.3.8   22-6-93 Fixed bug in consistp.c implementation of R(i,3) -- Phil Green
1.3.9	25-6-93 Finished gapdecayrate -- modified cutoffs()
1.3.10  7-7-93  Added -overlap2 option, prompted by Erik Sonnhammer
1.3.11	29-10-93	Added -dbrecmin and -dbrecmax options
*/
CharPtr	version = "1.3.11";
CharPtr	rel_date = "29-Oct-93";

int
main(argc, argv)
	int		argc;
	char	**argv;
{
	char	o = NULLB, *a;
	char	*theArg;
	int		i, ac, arglen;
	long	efflen;
	size_t	nheaders;
	Boolean	secondary;
	long	ilong;

	sys_times(&t0);
	alarmprocset();

	banner(argv[0], version, rel_date, citation, (char *)NULL);

	if (argc < 3)
		usage();

	for (ac = 3; ac < argc; ++ac) {
		theArg = argv[ac];
		arglen = strlen(theArg);
		if (str_casecmp("-over2", theArg) == 0 || str_casecmp("-overlap2", theArg) == 0) {
			ovlap = ovlap_a2;
			overlap_option = theArg;
			continue;
		}
		if (str_ncasecmp("-over", theArg, 5) == 0) {
			ovlap = ovlap_0;
			overlap_option = theArg;
			continue;
		}
		if (str_casecmp("-filter", theArg) == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing filter specification");
			filtercmd = pick_filter(argv[ac], QUERY_SEQTYPE);
			continue;
		}
		if (arglen >= 6 && arglen <= 11 &&
					str_ncasecmp("-echofilter", theArg, arglen) == 0) {
			echofilter_flag = TRUE;
			continue;
		}
		if (str_ncasecmp("-qoffset", theArg, 5) == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing -qoffset");
			if (sscanf(argv[ac], "%ld", &ilong) != 1)
				bfatal(ERR_INVAL, "Invalid value specified with -qoffset:  %s", argv[ac]);
			Qoffset = ilong;
			continue;
		}
		if (str_ncasecmp("-consis", theArg, 7) == 0) {
			consistency_flag = FALSE;
			continue;
		}
		if (str_ncasecmp(theArg, "-poisson", 8) == 0) {
			continue;
		}
		if (str_casecmp(theArg, "-gi") == 0) {
			gi_option = TRUE;
			continue;
		}
		if (str_casecmp(theArg, "-dbrecmax") == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing -dbrecmax value");
			if (sscanf(argv[ac], "%ld", &dbrecmax) != 1 || dbrecmax <= 0)
				bfatal(ERR_INVAL, "Invalid -dbrecmax value:  %s", argv[ac]);
			continue;
		}
		if (str_casecmp(theArg, "-dbrecmin") == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing -dbrecmin value");
			if (sscanf(argv[ac], "%ld", &dbrecmin) != 1 || dbrecmin <= 0)
				bfatal(ERR_INVAL, "Invalid -dbrecmin value:  %s", argv[ac]);
			--dbrecmin;
			continue;
		}
		if (str_casecmp("-sort_by_pvalue", theArg) == 0) {
			hitlist_cmp_func = cmp_by_pval;
			continue;
		}
		if (str_casecmp("-sort_by_count", theArg) == 0) {
			hitlist_cmp_func = cmp_by_count;
			continue;
		}
		if (str_casecmp("-sort_by_highscore", theArg) == 0) {
			hitlist_cmp_func = cmp_by_highscore;
			continue;
		}
		if (str_casecmp("-sort_by_totalscore", theArg) == 0) {
			hitlist_cmp_func = cmp_by_totalscore;
			continue;
		}
		if (str_ncasecmp("-gapdecayrate", theArg, 8) == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing gapdecayrate");
			if (sscanf(argv[ac], "%lg", &gapdecayrate) != 1)
				bfatal(ERR_INVAL, "Invalid value specified with -gapdecayrate:  %s", argv[ac]);
			continue;
		}
		secondary = FALSE;
		if (theArg[1] == '=')
			o = theArg[0];
		else
			if (theArg[0] == '-')
				o = theArg[1];
			else
				if (theArg[1] == '2' && theArg[2] == '=') {
					o = theArg[0];
					secondary = TRUE;
				}
				else
					bfatal(ERR_INVAL, "Argument %d (``%s'') was unexpected or has improper form.", ac, theArg);
		a = theArg + 2;
		if (secondary)
			++a;
		if (*a == NULLB) /* ignore */
			continue;
		if (islower(o))
			o = toupper(o);
		if (!secondary)
		switch (o) {
			case 'P':
#ifdef MPROC_AVAIL
				numprocs = atoi(a);
				if (numprocs <= 0)
					bfatal(ERR_INVAL, "Bad value for P (the number of processors)");
				if (numprocs > THREADS_MAX)
					warning("%s:  Pmax = %d.", module, THREADS_MAX);
				numprocs = MIN(numprocs, THREADS_MAX);
#else
				warning("Parameter P is ignored--this program has not been compiled to use multiprocessing");
#endif
				break;
			case 'M':
				if ((int)strlen(a) >= FILENAME_MAX)
					bfatal(ERR_INVAL, "PAM file name is too long");
				strcpy(M, a);
				if (strcmp(M, DEFAULT_M) != 0)
					M_set = TRUE;
				break;
			case 'W':
				if ((W = atoi(a)) <= 0)
					bfatal(ERR_INVAL, "Bad value for W, the word length.");
				if (W > WORDSIZE_MAX)
					bfatal(ERR_WORDSIZE, "W is too big--must be less than %d", WORDSIZE_MAX+1);
				W_set = TRUE;
				break;
			case 'S':
				if ((S = atol(a)) < 1)
					bfatal(ERR_INVAL, "Bad value for S, the cutoff score.");
				S_set = TRUE;
				break;
			case 'E':
				if (sscanf(a, "%lg", &E) != 1 || E <= 0.0 || E > DEFAULT_E_MAX)
					bfatal(ERR_INVAL, "Bad value for E, the expect value.");
				E_set = TRUE;
				break;
			case 'K':
				if (sscanf(a, "%lg", &K) != 1 || K < 0.)
					bfatal(ERR_INVAL, "Bad value for K, the Karlin-Altschul K parameter.");
				K_set = TRUE;
				break;
			case 'L':
				if (sscanf(a, "%lg", &Lambda) != 1 || Lambda <= 0.)
					bfatal(ERR_INVAL, "Bad value for L (Lambda)");
				L_set = TRUE;
				break;
			case 'V':
				if (sscanf(a, "%ld", &ilong) != 1 || (V = ilong) < 0)
					bfatal(ERR_INVAL, "Bad value for V, the number of descriptions.");
				V_set = TRUE;
				break;
			case 'T':
				if ((T = atoi(a)) < 1)
					bfatal(ERR_INVAL, "Bad value for T, the neighborhood word score threshold.");
				T_set = TRUE;
				break;
			case 'X':
				if ((X = atoi(a)) < 1 || X > 100000)
					bfatal(ERR_INVAL, "Bad value for X, the extension dropoff score.");
				X_set = TRUE;
				minusX = -X;
				break;
			case 'Y':
				{
					double	x;
					if (sscanf(a, "%lg", &x) != 1 || x > LONG_MAX)
						bfatal(ERR_INVAL, "Bad value for Y, the effective query sequence length.");
					Meff = x;
				}
				Y_set = TRUE;
				break;
			case 'Z':
				{
					double	x;
					if (sscanf(a, "%lg", &x) != 1 || x > LONG_MAX)
						bfatal(ERR_INVAL, "Bad value for Z, the effective database length.");
					Neff = x;
				}
				break;
			case 'B':
				if (sscanf(a, "%ld", &ilong) != 1 || (showblast = ilong) < 0)
					bfatal(ERR_INVAL, "Bad value for B, the max. number of alignments.");
				break;
			case 'H':
				showhist = atoi(a);
				break;
			default:
				usage();
		}
		else
		/* Secondary search options */
		switch (o) {
			case 'E':
				if (sscanf(a, "%lg", &E2) != 1 || E2 > DEFAULT_E_MAX || E2 < 0.)
					bfatal(ERR_INVAL, "Bad value for E2, the secondary expect value.");
				E2_set = TRUE;
				break;
			case 'S':
				if ((S2 = atol(a)) <= 0 || S2 > 1000000000)
					bfatal(ERR_INVAL, "Bad value for S2, the secondary cutoff score.");
				S2_set = TRUE;
				break;
			default:
				usage();
		}
	}

	if (showhist == 0 && showblast == 0 && V == 0)
		bfatal(ERR_INVAL, "All output is suppressed--no search will be performed");

	/* read file of substitution weights */
	get_weights(M, AAID_NAR, aa_atob, &alphasize, &minchar, &maxchar, &achars,
					&order, &lows, &highs, &maxcost, &matrx);

	get_query(argv[2]);
	wrap(stdout, "Query=  ", qtitle, qtitlelen, 79, qnamelen+1);
	printf("        (%s letters", lu2s(q_len));
	Qlen = q_len + Qoffset;
	if (Qoffset != 0)
		printf(", offset %s", ld2s(Qoffset));
	printf(")\n");

	if (seqfilter(filtercmd, aa_btoa, aa_atob,
				qtitle, qtitlelen,
				(UcharPtr)q_seq, q_len,
				(UcharPtr)q_seq, (size_t PNTR)&q_len) != 0)
		bfatal(ERR_UNDEF, "Non-zero return code from sequence filter");

	putc('\n', stdout);

	initdb(argv[1]);

	(void) fflush(stdout);

	/*
	It would be more accurate to call lambdak() for every sequence in
	the database, but then the program would run much slower.
	Another suggestion is to call lambdak() before reporting
	the probability of an HSP containing part of a specific database
	sequence.  For the sake of speed, though, lambdak() is called only once.
	*/
	if (lambdak(lows, highs, q_seq, q_len, &efflen, &Lambda, &K, &H) != 0)
		bfatal(ERR_UNDEF, "Could not calculate Lambda, K, and H");
	if (!Y_set)
		Meff = efflen;

	if (!X_set) {
		/* X = score associated with at least 10 bits of information */
		X = ceil(10./(Lambda/LN2));
		minusX = -X;
	}

	if (!T_set) {
		T = DEFAULT_T;
		if (W == 3) {
			/* Ad hoc equation for PAM matrices with scale = ln(2)/2 */
			T = ceil((4.676 + 1.3*log(H))/Lambda - 0.5);
			/*T = ceil((2.0 + 2.741*H)/Lambda - 0.5);*/
		}
		if (W == 4) {
			/* Ad hoc equation for PAM matrices with scale = ln(2)/2 */
			T = ceil((2.1 + 3.751*H)/Lambda - 0.5);
		}
		T = MAX(T, 1);
	}

	/* Determine the cutoff score, S */
	if (S_set && !E_set)
		E = 0.;
	cutoffs(&S, &E, &highs_exp, Lambda, K, H, Neff, Meff, gapdecayrate);

	/* Determine the secondary cutoff score, S2, to use */
	if (E2 == 0. && !S2_set)
		S2 = S;
	if ((E2_set && !S2_set && E2 == 0.) || (S2_set && S2 > S))
		E2 = 0., S2 = S;
	else {
		if (S2_set && !E2_set)
			E2 = 0.;
		/* Assume an "average" protein's length is 300 a.a. */
		cutoffs(&S2, &E2, &highs_dummy, Lambda, K, H, 300, 300, 0.);
		/* Adjust S2 to be in line with S, as necessary */
		S2 = MAX(S2, 1);
		if (S2 > S)
			S2 = S;
		E2 = stoe(S2, Lambda, K, H, 300, 300);
	}

	Ssave = S;
	S = S2;

	score_hist_dim = etos(0.01, Lambda, K, H, Neff, Meff) + 1;
	score_hist = (unsigned long *)ckalloc0(sizeof(score_hist[0])*(score_hist_dim+1));
	poolp = HSPPoolNew();
	hl_pool = HitListPoolNew();

	/* The following relation is safe to use, because should the score
	ever reach S during hit extension, the entire diagonal will be
	searched rigorously ignoring X anyway.
	*/
	X = MIN(X, S);
	minusX = -X;

	maxscore = fmaxscore(q_seq, q_len, maxcost);
	if (maxscore < Ssave)
		bfatal(ERR_SCORING, "The maximum achievable score with this query sequence is only %d, whereas the cutoff score for reporting HSPs is %d.  You might lower the cutoff score, raise the expect value (E) to at least %ld, or use a different scoring scheme.  The maximum expect value this program will permit to be set is %lg.",
			maxscore, Ssave,
			(long)gapdecay(1, stoe(maxscore, Lambda, K, H, Neff, Meff), gapdecayrate) + 1,
			DEFAULT_E_MAX);

	if (S2 < T)
		warning("The cutoff score S2 = %d is less than the neighborhood word score threshold, T = %d.", S2, T);

	sys_times(&t1);

	/* build table of critical (neighborhood) words */
	dp = bld_table_aa((DFAPtr)NULL, INT_MIN, q_seq, q_len, W, T, alphasize,
					&matrx, &maxcost, &order, &numwords, &numexact);
	end_table(dp);
	sys_times(&t2);

	RunWild("Searching", dbreccnt, dosearch);
	S = Ssave;

#ifdef SIGXCPU
	/* Report results without interruption */
	signal(SIGXCPU, SIG_IGN);
#endif

	dfanstates = dp->nstates;
	dfastatesize = dp->statesize;
	dfasize = dfa_size(dp);
	dfaextent = dfa_extent(dp);
	dfa_destruct(dp);

	collate_hits();

	hasHSP = hitlist_count(firsthl);
	sort_hitlists(&firsthl, hitlist_cmp_func);
	nheaders = MIN(hasHSP , MAX(V, showblast));
	if (nheaders < hasHSP)
		hitlist_truncate(&firsthl, nheaders);
	get_headers(DATABASE_SEQTYPE, firsthl);

	sys_times(&t3);

	print_hist(score_hist_dim, score_hist, S, T, Lambda, K, H,
						Meff, Neff, nbr_seqs);
	print_headers(firsthl, 1.0, FALSE, 80);
	print_HSPs(firsthl, 1.0, print_p);
	print_parms();
	print_stats();
	ckwarnings();
	exit_code(ERR_NONE);
	/*NOTREACHED*/
}

/* print_parms() -- print the parameters */
print_parms()
{
	(void) printf("\n\nParameters:\n");
	if (filtercmd != NULL && filtercmd[0] != NULLB)
		(void) printf("  Query filter:  \"%s\"\n", filtercmd);
	(void) printf("  E = %#0.2lg, S = %ld (%.1lf bits),  E2 = %#0.2lg, S2 = %s\n",
		E, (long)S, (double)S*Lambda/LN2,
		E2, (S2 != S ? lu2s(S2) : "not used")
		);
	(void) printf("  W = %d, T = %ld (%.1lf bits), X = %ld (%.1lf bits)\n",
		W, (long)T, (double)T*Lambda/LN2,
		(long)X, (double)X*Lambda/LN2);
	(void) printf("  M = %s\n", M);
	if (K_set) {
		printf("  K = %lg", K);
		if (L_set)
			printf(",");
	}
	if (L_set)
		printf("  Lambda = %lg\n", Lambda);
	else
		if (K_set)
			printf("\n");
	(void) printf("  H = %d, V = %d, B = %d\n", showhist, V, showblast);
	report_gapdecayrate();
	if (overlap_option != NULL)
		printf("  Overlap option:  %s\n", overlap_option);
	return 0;
}


/* print_stats() -- print the statistics of the search */
print_stats()
{
	(void) printf("\nStatistics:\n");
	(void) printf("  Lambda = %#0.3lg nats/unit score, Lambda/ln2 = %#0.3lg bits/unit score\n",
				Lambda, Lambda/LN2);
	(void) printf("  K = %#0.3lg, H = %#0.3lg bits/position\n", K, H/LN2);
	(void) printf("  Expected/Observed high score = %s (%.1lf bits) / %s (%.1lf bits)\n",
			ld2s(highs_exp), ((double)highs_exp)*(Lambda/LN2),
			ld2s(highs_obs), ((double)highs_obs)*(Lambda/LN2)
			);
	if (S > highs_exp && highs_obs == 0)
		(void) printf("  NOTE:  the cutoff score is greater than the expected high score\n");
	if (highs_exp > maxscore)
		(void) printf("  NOTE:  expected high score is greater than the max. achievable score %d\n",
				maxscore);
	(void) printf("  # of letters in query:  %s", lu2s(q_len));
	if (Y_set)
		(void) printf("  (Y = %s)\n", lu2s(Meff));
	else
		if (Meff != q_len)
			(void) printf("  (Effective length = %s)\n", lu2s(Meff));
		else
			(void) putchar('\n');
	(void) printf("  # of neighborhood words in query:  %s\n", lu2s(numwords));
	(void) printf("  # of exact words scoring below T:  %s\n", lu2s(numexact));
	(void) wrap(stdout, "  Database:  ", dbtitle, -1, 79, 13);
	(void) printf("  # of letters in database:  %s", lu2s(totdblen));
	if (Neff != totdblen)
		(void) printf("  (Z = %s)\n", lu2s(Neff));
	else
		(void) putchar('\n');
	(void) printf("  # of word hits against database:  %s\n", lu2s((long)numhits));
	(void) printf("  # of failed hit extensions:  %s\n", ld2s((long)numfailed));
	(void) printf("  # of excluded hits:  %s\n",
				ld2s((long)numhits - (long)numfailed - totHSP - overcnt));
	(void) printf("  # of successful extensions:  %s\n", ld2s((long)totHSP+overcnt));
	(void) printf("  # of overlapping HSPs discarded:  %s\n",
				lu2s((long)overcnt));
	(void) printf("  # of HSPs reportable:  %s\n", ld2s((long)totHSP));
	(void) printf("  # of sequences in database:  %s\n", ld2s((long)nbr_seqs));
	(void) printf("  # of database sequences satisfying E:  %s\n",
				lu2s((long)hasHSP));

	(void) printf("No. of states in DFA:  %s (%s KB)\n",
			lu2s(dfanstates),
			lu2s(HOWMANY(dfanstates*dfastatesize, KBYTE)));
	(void) printf("Total size of DFA:  %s KB (%s KB)\n",
			lu2s(HOWMANY(dfasize, KBYTE)),
			lu2s(HOWMANY(dfaextent, KBYTE)));
	sys_showtime("Time to generate neighborhood", &t1, &t2, stdout);
#ifdef MPROC_AVAIL
	(void) printf("No. of processors used:  %d\n", numprocs);
#endif
	sys_showtime("Time to search database", &t2, &t3, stdout);
	sys_times(&t3);
	sys_showtime("Total cpu time", (Cputime_t *)NULL, &t3, stdout);
}

unsigned	hashsp;

static void
dosearch(userp, tp)
	Nlm_VoidPtr	userp;
	TaskBlkPtr	tp;
{
	register HSPPtr	hp;
	int		hashsp_local = 0, rem_max_per_proc;
	Boolean	saved = FALSE;
	MPDEF

	rem_max_per_proc = REMEMBER_MAX / tp->proc_max;
#ifdef MPROC_AVAIL
	initmpc(tp->id, mp = &mpc[tp->id]);
#endif
	/* For each sequence in the database */
	while (TaskNext(tp) >= 0) {
		get_seq(MPC tp->task_cur + dbrecmin);
		MPPTR high_score = 0;

		/* search the database sequence */
		search_aa(MP);

		/* keep track of the highest score */
		if (MPPTR high_score < score_hist_dim)
			++MPPTR score_hist[MPPTR high_score];
		else
			++MPPTR score_hist[score_hist_dim];
		if (MPPTR high_score > MPPTR highs_obs)
			MPPTR highs_obs = MPPTR high_score;

		/* Save any HSPs that were found */
		if (MPPTR curhl.hp != NULL) {
			sort_HSPs(&MPPTR curhl);
			if (consistency_flag) {
				consistn(&MPPTR curhl);
				consist_evals(&MPPTR curhl, Meff, MPPTR s_len, Neff);
			}
			else {
				pcnt(&MPPTR curhl);
				evals(&MPPTR curhl, Meff, MPPTR s_len, Neff);
			}
			(void) hitlist_prune_by_cutoff(&MPPTR curhl, MPPTR poolp, SCORET_EXPECT, E);
			if (MPPTR curhl.hp == NULL)
				continue;
			for (hp = MPPTR curhl.hp; hp != NULL; hp = hp->next)
				hp->P_val = etop(hp->expect);
			MPPTR curhl.seq_nbr = tp->task_cur + dbrecmin;
			if (++hashsp_local + hashsp > rem_max_per_proc) {
				M_LOCK;
				if (!saved) {
					saved = TRUE;
					hashsp += hashsp_local - 1;
					hashsp_local = 0;
				}
				if (++hashsp > REMEMBER_MAX)
					bfatal(ERR_REMEMBER,
					"More than %u sequences have HSPs to remember; increase REMEMBER_MAX and recompile.",
					REMEMBER_MAX);
				M_UNLOCK;
			}
			hitlist_save(MPPTR hl_pool, &MPPTR firsthl, &MPPTR curhl);
			MPPTR totHSP += MPPTR curhl.hspcnt;
			MPPTR curhl.hp = NULL;
			MPPTR curhl.hspcnt = 0;
		}
	}
	M_LOCK;
	hashsp += hashsp_local;
	M_UNLOCK;
}

#ifdef MPROC_AVAIL
/* Initialize the processor context pointed to by mp */
initmpc(id, mp)
    int     id;
    MPContextPtr    mp;
{    
    int     len;
    register int    i;
 
    mp->id = id;
 
    mp->nbr_seqs = -1;
 
    len = maxlen + q_len + 1;
    mp->_dlev = (DIAG_TYPE *)ckalloc0(sizeof(*_dlev)*len);
    mp->diag_level = mp->_dlev + q_len;
    mp->diag_top = mp->diag_stack = (DIAG_PTR)ckalloc(sizeof(*diag_top)*len);
    mp->s_seq0 = (char *)ckalloc(maxlen+2);
    *mp->s_seq0++ = NULLB;
    mp->s_seq = mp->s_seq0;
	if (id == 0) {
		mp->score_hist = score_hist;
		mp->poolp = poolp;
		mp->hl_pool = hl_pool;
	}
	else {
		mp->score_hist
				= (unsigned long *)ckalloc0(sizeof(mp->score_hist[0])*(score_hist_dim+1));
		mp->poolp = HSPPoolNew();
		mp->hl_pool = HitListPoolNew();
	}
}
#endif

/* get_query - read the query sequence */
get_query(filename)
	char *filename;
{
	char	*base, *bp;
	size_t	basemax;
	FILE	*fp;
	register char *p, ch;
		
	fp = ckopen(filename, "r", TRUE);

	base = ckalloc(basemax = 10*KBYTE);
	bp = base + 1;

	while (vfgets(&base, &basemax, &bp, fp) != NULL) {
		if (*bp == '>') {	/* Optional header/title line */
			if (bp != base+1)/* Second sequence encountered */
				break;
			p = str_chr(bp, '\n');
			if (p != NULL)
				*p = NULLB;
			qtitle = str_dup(bp+1);
			qtitlelen = strlen(qtitle);
			for (p=qtitle; (ch = *p) && !isspace(ch); ++p)
				;
			*p = NULLB;
			qname = str_dup(qtitle);
			*p = ch;
			qnamelen = strlen(qname);
			continue;
		}
		p = bp;
		while ((ch = *p++) != NULLB) {
			if (aa_atob[ch] <= AAID_MAX) {
				if (islower(ch))
					ch = toupper(ch);
				*bp++ = ch;
			}
			else
				if (aa_atob[ch] == AAID_NAR) {
					aa_atob[ch] = AAID_IGNORE;
warning("Invalid amino acid code ``%c'' encountered in query sequence will be skipped.", ch);
				}
		}
		if (bp - base > QUERYLEN_MAX-1)
			bfatal(ERR_QUERYLEN, "query sequence exceeds %d characters in length.",
				QUERYLEN_MAX-1);
			
	}
	(void) fclose(fp);
	_q_seq = base;
	_q_seq[0] = NULLB;
	q_seq = base+1;
	q_len = bp - q_seq;
	if (q_len < W)
		bfatal(ERR_QUERYLEN, "query sequence is shorter than the word length, W=%d.", W);
	*bp = NULLB;

	(void) ckseqtype(QUERY_SEQTYPE, q_seq, q_len);

	p = q_seq;
	while ((ch = *p) != NULLB)
		*p++ = aa_atob[ch];

	qq_seq = ckalloc(q_len+1);
	return 0;
}

/* UNIX-specific input routine; database must be pre-processed by SETDB */

static void
initdb(dbname)
	char	*dbname;
{
	struct stat	sbuf;
	MFILE	*tfile;
	long	dbtype, dbformat;
	char	*dirname = "";
	char	name[FILENAME_MAX];
	long	i;
	long	many, len;
	int		once=0;

RETRY:
	if (*dirname != NULLB)
		(void) sprintf(dbfname, "%s%s%s", dirname, DIRDELIMSTR, dbname);
	else
		SAFENCPY(dbfname, dbname, sizeof(dbfname));

	(void) sprintf(name, "%s%s", dbfname, TABLE_EXT);
	tfile = mfil_open(name, "r", MFIL_OPT_ALL);
	if (tfile == NULL) {
		if ((dirname = getenv("BLASTDB")) == NULL)
			dirname = BLASTDB;
		if (++once == 1)
			goto RETRY;
		bfatal(ERR_MFILE, "couldn't open database file ``%s''.  Make sure you have specified the correct database name and requested a database of the correct type (protein) for use with this program.", name);
	}
	if (mfil_dup_long(tfile, &dbtype, 1) != 1)
		bfatal(ERR_MFILE, "Read error on file %s", name);
	if (dbtype != DB_TYPE_PRO)
		bfatal(ERR_DBASE, "``%s'' is not a protein sequence database.", dbname);
	if (mfil_dup_long(tfile, &dbformat, 1) != 1)
		bfatal(ERR_DBASE, "Corrupted database file:  %s", name);
	if (dbformat != AAFORMAT)
		bfatal(ERR_DBASE, "Database format is an incompatible version; rerun ``setdb''.");

	if (mfil_dup_long(tfile, &i, 1) != 1)
		bfatal(ERR_DBASE, "Corrupted database file:  %s", name);
	if ((dbtitle = mfil_get(tfile, i)) == NULL)
		bfatal(ERR_DBASE, "Corrupted database file:  %s", name);
	if (i%4 != 0) {
		i = 4 - (i%4);
		mfil_seek(tfile, i, SEEK_CUR);
	}

	if (mfil_dup_long(tfile, (long *)&nbr_seqs, 1) != 1
		|| mfil_dup_long(tfile, (long *)&totdblen, 1) != 1
		|| mfil_dup_long(tfile, (long *)&maxlen, 1) != 1 )
			bfatal(ERR_DBASE, "corrupted database file:  %s", name);

	if (dbtitle[0] == NULLB)
		dbtitle = str_dup(dbname);
	(void) wrap(stdout, "Database:  ", dbtitle, -1, 79, 11);
	(void) printf("           %s sequences; %s total letters.\n",
			ld2s(nbr_seqs), ld2s(totdblen));

	if (nbr_seqs <= 0 || totdblen <= 0)
		bfatal(ERR_DBASE, "Nothing in the database to search!?");

	ss_seq = (char *)ckalloc(sizeof(*ss_seq)*(maxlen+1));

	many = nbr_seqs + 1;
	seq_beg = (unsigned long *)mfil_get_long(tfile, many);
	header_beg = (unsigned long *)mfil_get_long(tfile, many);
	if (seq_beg == NULL || header_beg == NULL)
		bfatal(ERR_DBASE, "improper offsets in table file.");

	(void) sprintf(name, "%s%s", dbfname, AA_HEADER_EXT);
	hfile = mfil_open(name, "r", MFIL_OPT_ALL);
	if (hfile == NULL)
		bfatal(ERR_FOPEN, "Couldn't open database file %s", name);

	(void) sprintf(name, "%s%s", dbfname, AA_SEARCHSEQ_EXT);
	/* Attach a shared memory segment for the .bsq file (if one exists) */
	sfile = mfil_open(name, "r", MFIL_OPT_ALL);
	if (sfile == NULL)
		bfatal(ERR_FOPEN, "Couldn't open database file %s", name);
	if ((dbbuf = MFILE_SP(sfile)) == NULL) {
#ifndef MPROC_AVAIL
		(void) mfil_seek(sfile, (long)seq_beg[0], SEEK_SET);
		s_seq = (char *)ckalloc(sizeof(*s_seq)*(maxlen+2));
		*s_seq++ = NULLB;
#else
		long	dbsize;
		fflush(stdout);
		fstat(fileno(MFILE_FP(sfile)), &sbuf);
		dbsize = sbuf.st_size;
		s_seq0 = (char *)ckalloc(sizeof(*s_seq0)*dbsize);
		if (mfil_read((char *)s_seq0, (int)dbsize, 1, sfile) != 1)
			bfatal(ERR_DBASE, "Database read error.");
#endif /* MPROC_AVAIL */
	}

	len = maxlen + q_len + 1;
	_dlev = (DIAG_TYPE *)ckalloc0(sizeof(*_dlev)*len);
	diag_level = _dlev + q_len;

	diag_top = diag_stack = (DIAG_PTR)ckalloc(sizeof(*diag_top)*len);

	if (Neff == 0)
		Neff = totdblen;

	dbrecmin = MAX(dbrecmin, 0);
	dbrecmin = MIN(dbrecmin, nbr_seqs);
	dbrecmax = (dbrecmax > 0 ? MIN(dbrecmax, nbr_seqs) : nbr_seqs);
	dbreccnt = dbrecmax - dbrecmin;
	if (dbreccnt < nbr_seqs)
		printf("           Subset of database from %ld to %ld.\n",
			dbrecmin+1, dbrecmax);
}

/* get_seq - read a database sequence and header line, return nonzero on EOF */
static int
get_seq(MPC current)
	MPDEF
	unsigned long	current;
{
	if (dbbuf != NULL) {
		MPPTR s_len = seq_beg[current+1] - seq_beg[current] - 1;
		MPPTR s_seq = dbbuf + seq_beg[current];
	}
	else {
#ifndef MPROC_AVAIL
		s_len = seq_beg[current+1] - seq_beg[current];
		if (mfil_read(s_seq, s_len, 1, sfile) != 1)
			bfatal(ERR_DBASE, "error reading database sequence.");
		--s_len;
#else
		MPPTR s_len = seq_beg[current+1] - seq_beg[current] - 1;
		MPPTR s_seq = s_seq0 + seq_beg[current];
#endif
	}
	if (MPPTR s_len > maxlen)
		bfatal(ERR_DBASE, "Database sequence longer than maxlen!  Sequence #%d.", current);
	return 0;
}

/*****************************/
#include "searcha.inc"
/*****************************/

/* ----------------------------  parameter calculations  ---------------------*/

/* Version of lambdak() specifically for a.a. alphabet used by BLASTP */
int
lambdak(lows, highs, q_seq, q_len, efflen, Lambda, K, H)
	long	lows, highs;
	unsigned long	q_len;
	long	PNTR efflen;
	register char	*q_seq;
	double	*Lambda, *K, *H;
{
	double	fq1[AAID_MAX+1];
	unsigned long	iq1[AAID_MAX+1];
	double	*prob;
	unsigned long	range, qlen;
	register unsigned long	ui;
	register int	i, j;

	if (lows >= 0 || highs <= 0)
		return 1;

	/* Tally the frequency of occurrence of each letter in the query */
	for (i=0; i<DIM(iq1); ++i)
		iq1[i]=0;
	for (ui=0, qlen = q_len; ui<q_len; ++ui) {
		/* only count unambiguous letters */
		if (q_seq[ui] != aa_atob[UNKNOWN_AA_CHR])
			++iq1[q_seq[ui]];
		else
			--qlen;
	}
	*efflen = qlen;

	/* Convert frequencies to probabilities */
	for (i=AAID_MIN; i<=AAID_MAX; ++i)
		fq1[i] = (double)iq1[i] / qlen;

	/* Calculate probability of each score */
	range = highs - lows;
	prob = (double *)ckalloc(sizeof(*prob)*(range+1));
	for (ui=0; ui<=range; ++ui)
		prob[ui] = 0.;

	for (i=minchar; i<=maxchar; ++i)
		for (j=AAID_MIN; j<DIM(aafq); ++j) {
			if (matrx[i][j] > SCORE_MIN/WORDSIZE_MAX)
				prob[matrx[i][j]-lows] += fq1[i]*aafq[j];
		}

	i = karlin(lows, highs, prob, !L_set, Lambda, !K_set, K, TRUE, H);
	mem_free((char *)prob);
	if (i != 0) {
		*K = *Lambda = -1.;
		return i;
	}
	return 0;
}

static int
print_p(hlp)
	HitListPtr	hlp;
{
	Seqid_t	current;
	Coord_t	s_len;
	HSPPtr	hp;
	int		i;
	register char	*cp, ch;

	current = hlp->seq_nbr;

	cp = hlp->header;
	while ((ch = *cp) != NULLB && !isspace(ch))
		++cp;
	while ((ch = *cp) != NULLB && isspace(ch))
		++cp;
	i = cp - hlp->header + 1;
	i = MIN(i, 12);
	putchar('\n');
	putchar('\n');
	(void) wrap(stdout, ">", hlp->header, hlp->header_len, 79, i);

	/* get sequence */
	if (dbbuf != NULL) {
		s_len = seq_beg[current+1] - seq_beg[current] - 1;
		s_seq = dbbuf + seq_beg[current];
	}
	else {
#ifndef MPROC_AVAIL
		s_len = seq_beg[current+1] - seq_beg[current];
		if (mfil_seek(sfile, seq_beg[current], SEEK_SET) == -1)
			bfatal(ERR_DBASE, "fseek error on database sequence file.");
		if (mfil_read(s_seq, s_len, 1, sfile) != 1)
			bfatal(ERR_DBASE, "error reading database sequence.");
		--s_len;
#else
		s_seq = s_seq0 + seq_beg[current];
		s_len = seq_beg[current+1] - seq_beg[current] - 1;
#endif
	}
	(void) printf("%*sLength = %s\n", i, "", ld2s((long)s_len));
	
	for (hp = hlp->hp; hp != NULL; hp = hp->next) {

		if (hsp_print(stdout, hp, "", (UcharPtr)q_seq + hp->q_pos, (UcharPtr)s_seq + hp->s_pos, matrx, aa_btoa, aa_btoa, Qlen, s_len, nox_coord, nox_coord, nox_coord, nox_coord, NULL, NULL, Lambda) != 0)
			return -1;
	}
	return 0;
}

usage()
{
	fflush(stdout);
	fflush(stderr);
	fprintf(stderr, "\nUsage:\n\n");
	fprintf(stderr,
		"    %s database queryfile [options]\n\n", module);
	fprintf(stderr,
		"Valid %s options:  E, S, E2, S2, W, T, X, M, Y, Z, K, L, H, V, and B\n",
		module);
	fprintf(stderr, "\t-qoffset #\n");
	fprintf(stderr, "\t-dbrecmin #\n");
	fprintf(stderr, "\t-dbrecmax #\n");
	fprintf(stderr, "\t-gi (gi identifier suppression)\n");
	fprintf(stderr, "\t-overlap\n\t-overlap2\n\t-gapdecayrate #\n\t-filter filteroption\n\t-echofilter\n\t-consistency\n");
	fprintf(stderr, "\t-sort_by_pvalue\n\t-sort_by_count\n\t-sort_by_highscore\n\t-sort_by_totalscore\n");
	exit_code(ERR_INVAL);
	/*NOTREACHED*/
}

bad_num_arg(c, arg)
	int	c;
	CharPtr	arg;
{
	bfatal(ERR_INVAL, "Non-numeric argument or bad value for parameter %c:  %s", c, arg);
	/*NOTREACHED*/
}
