/**************************************************************************
*                                                                         *
*                             COPYRIGHT NOTICE                            *
*                                                                         *
* This software/database is categorized as "United States Government      *
* Work" under the terms of the United States Copyright Act.  It was       *
* produced as part of the author's official duties as a Government        *
* employee and thus can not be copyrighted.  This software/database is    *
* freely available to the public for use without a copyright notice.      *
* Restrictions can not be placed on its present or future use.            *
*                                                                         *
* Although all reasonable efforts have been taken to ensure the accuracy  *
* and reliability of the software and data, the National Library of       *
* Medicine (NLM) and the U.S. Government do not and can not warrant the   *
* performance or results that may be obtained by using this software,     *
* data, or derivative works thereof.  The NLM and the U.S. Government     *
* disclaim any and all warranties, expressed or implied, as to the        *
* performance, merchantability or fitness for any particular purpose or   *
* use.                                                                    *
*                                                                         *
* In any work or product derived from this material, proper attribution   *
* of the author(s) as the source of the software or data would be         *
* appreciated.                                                            *
*                                                                         *
**************************************************************************/
/*
TBLASTN - Search a nucleotide database for regions that approximately match
regions of a protein query sequence.  The basic command syntax is

 	TBLASTN database query_sequence

	query_sequence names a file containing a protein sequence (where the
	> header line is optional).

The TBLASTN command permits optional additional arguments, such as X=50,
that reset certain internal parameters.  The available parameters are:

	E or e is the _expected_ no. of HSPs to produce (which influences S)
		(default is 25).
	S or s gives the High-scoring Segment Pair (HSP) cutoff score
		(default is calculated from E).
	T or t gives the threshold for a word match (default is calculated,
		but is usually about 15 with the PAM120 matrix).
	X or x gives the value for terminating word extensions (default is 20)
	M or m is the name of a file containing substitution costs matrix
		(default is "BLOSUM62")
	Y or y is the effective length of the querY sequence for statistical
		significance calculations (default uses the actual query length)
	Z or z is the effective databaZe length for statistical significance
		calculations (default uses the actual database length)
	C or c is the identifying number of the genetic code to use.  The
		default value is 0, which represents the standard genetic code.
		See gcode.h for the list of pre-defined genetic codes.
	P or p is the number of processors to utilize when the program
		is compiled to run on a multi-processor computer.

Thus a typical command is

	tblastn genbank myseq e=0.5 x=25

Literature:

    Karlin & Altschul, "Methods for assessing the statistical significance
        of molecular sequence features by using general scoring schemes,"
        Proc. Natl. Acad. Sci. USA 87 (1990) 2264-2268;
    Altschul, Gish, Miller, Myers and Lipman, "Basic local alignment search
        tool," J. Mol. Biol. 214 (1990) in press.

 (Note:  set tabstops every 4 columns for a better appearance).
*/
#include <ncbi.h>
#include <gish.h>
#include <gishlib.h>

#define EXTERN
#ifndef TBLASTN
#define TBLASTN
#endif
#include "blastapp.h"
#include "aabet.h"	/* the alphabet */
#include "ntbet.h"
#include "gcode.h"

#define CDBLEN_MAX	60000000	/* max. length of compressed database */

#define DEFAULT_M	"BLOSUM62"	/* default scoring matrix */
#define DEFAULT_T	13	/* default threshold for neighborhood word hits */
#define DEFAULT_X	20	/* default score drop-off during hit extensions */
#define DEFAULT_E2	0.15	/* default secondary Expect cutoff */
#define DEFAULT_S2	0	/* default secondary cutoff score */
#define DEFAULT_W	3	/* default word size */

char
	dbfname[FILENAME_MAX],
	name[FILENAME_MAX],
	*qname = "NO NAME",
	*qtitle = "NO NAME",
	*dbtitle,
	*dbbuf,			/* Memory-resident database */
	*d_seq,
	*s_seq0, *s_seq1, *s_seq2, *s_seq, *ss_seq,
	*_q_seq,	/* null prefix + query sequence + null terminator */
	*q_seq,		/* = _q_seq+1 = query sequence + null terminator */
	*qq_seq,	/* binary-encoded query sequence + null terminator */
	M[FILENAME_MAX] = DEFAULT_M;	/* file containing the substitution scores */

int		qnamelen = 7, qtitlelen = 7;

unsigned long
		totdblen;	/* total no. of letters in the database */

unsigned long	line_len,
		clean_count,
		*cseq_beg;

long	dbtype, dbformat;

unsigned long
	maxlen,		/* max length of a database sequence */
	numwords,	/* number of critical words in hash table */
	numexact,	/* number of exact words that did not score at least T */
	numbogus;	/* number of alignments with X, or when number is unknown */

Score_t
	highs_dummy;

Score_t
	S = DEFAULT_S,		/* cutoff score for High-scoring Segment Pairs (HSP) */
	S2 = DEFAULT_S2;	/* secondary cutoff score once at least 1 HSP found */

int
	S_set,
	S2_set,
	T = DEFAULT_T,		/* threshold for word hits */
	T_set,
	X = DEFAULT_X,		/* constant for word-hit extensions */
	minusX = -DEFAULT_X, /* -X */
	X_set,
	Y_set,
	W = DEFAULT_W,		/* word size */
	W_set,
	C = DEFAULT_C,		/* genetic code */
	C_set,
	E_set,
	E2_set,
	M_set;

double
	E = DEFAULT_E,		/* Expected number of HSPs achieving cutoff score */
	E2 = DEFAULT_E2;		/* Secondary Expect value */

DIAG_TYPE
	*_dlev,		/* flags for which diagonals... */
	*diag_level;	/* have been rigorously examined */

DIAG_PTR
	diag_stack,	/* list of diagonals with matches */
	diag_top;	/* next diagonal with a match */

static Score_t	maxscore;	/* maximum achievable score */

static FILE
	*sfile;			/* file pointer for sequences */

static MFILE
	*cfile,			/* file pointer for compressed sequences */
	*tfile;			/* memory file pointer for table */

static Score_t
	highs,		/* highest score in the PAM matrix */
	lows;		/* lowest score in the PAM matrix */

unsigned long	*seq_beg; /* starts of sequences in file or memory-resident dbase */

static char
	*ambiguity_ray;

DFAPtr	dp;	/* automaton for neighborhood word matching */

int		job_strands = TOP_STRAND|BOTTOM_STRAND;
int		strand;

int		alphasize;			/* alpabet size */
int		minchar, maxchar;	/* min. and max. letter values */
static char
	/* array of amino acid names, plus null terminator */
	achars[ALPHASIZE_MAX+1],
	/* order[i][j] = j-th best char to sub. for i */
	order[ALPHAVAL_MAX+1][ALPHASIZE_MAX];

unsigned char	gcode[64], revgcode[64];	/* genetic codes */
int		frame;	/* current reading frame */

Score_t
	/* max cost of substituting for a character */
	maxcost[ALPHAVAL_MAX+1],
	/* matrx[i][j] = cost of replacing i by j */
	matrx[ALPHAVAL_MAX+1][ALPHAVAL_MAX+1];

Nlm_Boolean	consistency_flag = TRUE;

static void	search_aa PROTO((MPPROTO));
static void	trans2 PROTO((CharPtr ntq, CharPtr aaq, int len));
static void rtrans2 PROTO((CharPtr ntq, CharPtr aaq, int len));
static void	trans2b PROTO((CharPtr ntq, CharPtr aaq, int len));
static void rtrans2b PROTO((CharPtr ntq, CharPtr aaq, int len));
static void	initdb PROTO((char *dbname));
static int	get_seq PROTO((MPPROTOC unsigned long));
static void	dosearch PROTO((Nlm_VoidPtr userp, TaskBlkPtr tp));
static int	print_t PROTO((HitListPtr));

unsigned long	dfanstates, dfastatesize, dfasize, dfaextent;

Cputime_t	t0, t1, t2, t3;

char	*citation = "Altschul, Stephen F., Warren Gish, Webb Miller, Eugene W. Myers, and David J. Lipman (1990).  Basic local alignment search tool.  J. Mol. Biol. 215:403-410.";

char	*notice = "statistical significance is estimated under the assumption that the equivalent of one complete reading frame of the database codes for protein and that significant alignments will involve only coding reading frames.";

/*
1.1.5	11-24-90	When at least one HSP is found involving a given database
					sequence, it is scanned again using a lower cutoff score.
					The amount the cutoff score is reduced is the V parameter.
1.1.6	12-4-90		Fixed BLAST algorithm to halt hit extensions properly
1.1.7	12-6-90		Moved many statistical calculations into library functions
1.1.8	12-7-90		Using the improved DFA library code
1.1.9	12-15-90	Using more common code, new hit lists
1.1.10	1-14-91		Fixed bug in bldaa.c regarding perfect matches
1.1.11	1-16-91		Fixed calculation of Poisson P-values
1.1.12	1-17-91		The longest sequence in the database was not being handled
					properly:  null terminators of the translated reading
					frames were being overwritten.
1.1.13	1-28-91		Display one-line description of each db sequence hit
1.1.14	2-4-91		Word-wrap long sequence descriptions
1.1.15	2-25-91		Added "Searching..." progress indicator
1.1.16	3-13-91		Added calculation of T when W=3
1.1.18	22-7-91	Default X is calculated using information theory
1.1.19	13-9-91	Changed behavior of parameters V and B.
1.1.20	15-9-91	Specific amino acids are properly inferred from ambiguous codons
1.1.21	23-9-91	Improved efficiency in search_aa() and extend_aa().
1.1.21	25-9-91	Improved reporting of individual HSP statistics
1.2.0	1-10-91	Improved Poisson statistics, screening and sorting of hits
1.2.1	23-12-91	Improved command line parsing, new -overlap option
1.2.2	26-12-91	Improved error reporting; new K & L options
1.2.3	6-1-92	Only use unambiguous letters (non-X) when calculating K & L
1.2.4	18-2-92	Switched to new dfa library
1.2.5	9-3-92	Faster K calculation in karlin()
1.2.6	28-3-92	Independence from the original FASTA-format database file
1.2.7	31-3-92	Fixed multiprocessing bug in translate() that failed to set
				s_len for reading frame 1.
				Added gap character '-' to amino acid alphabet.
1.2.8	12-6-92	Fixed bug in lambdak() when smaller alphabets are used
1.2.9	16-6-92	Added several hitlist sorting options
1.2.10	25-8-92	DEC Ultrix compatibility
1.2.11	4-9-92	Fixed bug in lib/hsppool.c
1.2.12	26-10-92  Fixed bug in segmented sequence handling
1.3.0   22-11-92 Added consistent Poisson Event count determination
1.3.1   23-4-93 Added consistent Poisson P-value calculation
1.3.2   5-5-93  Karlin H was assumed to have wrong units in lib/stolen.c
1.3.3   9-5-93  Improved handling of database sequences w/ambiguity codes.
                Lowered primary cutoff score in search_aa() to be consistent
                with BLASTP; TBLASTN is thus significantly slower but more
                sensitive.
1.3.4	9-6-93	Added gapdecayrate suggested by Phil Green; fixed overflow bug
                in lib/consistp.c
1.3.5	21-6-93	Fixed bug in consistent N count calculation for (-) strand hits
1.3.6   22-6-93 Fixed bug in consistp.c implementation of R(i,3) -- Phil Green
1.3.7   25-6-93 Finished gapdecayrate -- modified cutoffs()
1.3.8	7-7-93	Added -overlap2 option, prompted by Erik Sonnhammer
1.3.9	29-10-93	Added -dbrecmin and -dbrecmax options
1.3.10	30-6-94	Updated the genetic codes
1.3.11	25-8-94	Fixed db translation correction vs. nh words bug in searcha.inc
*/
CharPtr	version = "1.3.11";
CharPtr	rel_date = "25-Aug-94";

int
main(argc, argv)
	int		argc;
	char	**argv;
{
	char	o = NULLB, *a, *theArg;
	int		i, ac, arglen;
	long	efflen;
	size_t	nheaders;
	Boolean	secondary;

	sys_times(&t0);
	alarmprocset();

	banner(argv[0], version, rel_date, citation, notice);

	if (argc < 3)
		usage();

	for (ac=3; ac < argc; ++ac) {
		arglen = strlen(theArg = argv[ac]);
		if (str_casecmp("-over2", theArg) == 0 || str_casecmp("-overlap2", theArg) == 0) {
			ovlap = ovlap_t2;
			overlap_option = theArg;
			continue;
		}
		if (arglen >= 5 && str_ncasecmp(theArg, "-over", 5) == 0) {
			ovlap = ovlap_0;
			overlap_option = theArg;
			continue;
		}
		if (arglen >= 7 && str_ncasecmp("-filter", theArg, 7) == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing filter specification");
			filtercmd = pick_filter(argv[ac], QUERY_SEQTYPE);
			continue;
		}
		if (arglen >= 7 && str_ncasecmp("-consis", theArg, 7) == 0) {
			consistency_flag = FALSE;
			continue;
		}
		if (str_ncasecmp(theArg, "-poisson", 8) == 0) {
			continue;
		}
		if (str_casecmp(theArg, "-gi") == 0) {
			gi_option = TRUE;
			continue;
		}
		if (arglen >= 6 && arglen <= 11 &&
				str_ncasecmp("-echofilter", theArg, arglen) == 0) {
			echofilter_flag = TRUE;
			continue;
		}
		if (str_casecmp(theArg, "-dbrecmax") == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing -dbrecmax value");
			if (sscanf(argv[ac], "%ld", &dbrecmax) != 1 || dbrecmax <= 0)
				bfatal(ERR_INVAL, "Invalid -dbrecmax value:  %s", argv[ac]);
			continue;
		}
		if (str_casecmp(theArg, "-dbrecmin") == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing -dbrecmin value");
			if (sscanf(argv[ac], "%ld", &dbrecmin) != 1 || dbrecmin <= 0)
				bfatal(ERR_INVAL, "Invalid -dbrecmin value:  %s", argv[ac]);
			--dbrecmin;
			continue;
		}
		if (str_casecmp(theArg, "top") == 0) {
			job_strands &= (0xffffffff - BOTTOM_STRAND);
			continue;
		}
		if (str_casecmp(theArg, "bottom") == 0) {
			job_strands &= (0xffffffff - TOP_STRAND);
			continue;
		}
		if (str_ncasecmp("-qoffset", theArg, 5) == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing -qoffset");
			if (sscanf(argv[ac], "%ld", &Qoffset) != 1)
				bfatal(ERR_INVAL, "Invalid value specified with -qoffset:  %s", argv[ac]);
			continue;
		}
		if (str_casecmp("-sort_by_pvalue", theArg) == 0) {
			hitlist_cmp_func = cmp_by_pval;
			continue;
		}
		if (str_casecmp("-sort_by_count", theArg) == 0) {
			hitlist_cmp_func = cmp_by_count;
			continue;
		}
		if (str_casecmp("-sort_by_highscore", theArg) == 0) {
			hitlist_cmp_func = cmp_by_highscore;
			continue;
		}
		if (str_casecmp("-sort_by_totalscore", theArg) == 0) {
			hitlist_cmp_func = cmp_by_totalscore;
			continue;
		}
		if (str_ncasecmp("-gapdecayrate", theArg, 8) == 0) {
			if (++ac >= argc)
				bfatal(ERR_INVAL, "Missing gapdecayrate");
			if (sscanf(argv[ac], "%lg", &gapdecayrate) != 1)
				bfatal(ERR_INVAL, "Invalid value specified with -gapdecayrate:  %s", argv[ac]);
			continue;
		}
		secondary = FALSE;
		if (theArg[1] == '=')
			o = theArg[0];
		else
			if (*theArg == '-')
				o = theArg[1];
			else
				if (theArg[1] == '2' && theArg[2] == '=') {
					o = theArg[0];
					secondary = TRUE;
				}
				else
					bfatal(ERR_INVAL, "Argument %d (``%s'') was unexpected or has improper form.", ac, theArg);
		a = theArg + 2;
		if (*a == NULLB) /* ignore */
			continue;
		if (secondary)
			++a;
		if (islower(o))
			o = toupper(o);
		switch (o) {
			case 'P':
#ifdef MPROC_AVAIL
				numprocs = atoi(a);
				if (numprocs <= 0)
					bfatal(ERR_INVAL, "Bad value for P (the number of processors)");
				if (numprocs > THREADS_MAX)
					warning("%s:  Pmax = %d.", module, THREADS_MAX);
				numprocs = MIN(numprocs, THREADS_MAX);
#else
				warning("Parameter P is ignored--this program has not been compiled to use multiprocessing.",
						module);
#endif
				break;
			case 'M':
				if ((int)strlen(a) >= FILENAME_MAX)
					bfatal(ERR_INVAL, "PAM file name is too long");
				strcpy(M, a);
				if (strcmp(M, DEFAULT_M) != 0)
					M_set = TRUE;
				break;
			case 'W':
				if ((W = atoi(a)) <= 0)
					bfatal(ERR_INVAL, "Bad value for W");
				if (W > WORDSIZE_MAX)
					bfatal(ERR_WORDSIZE, "W is too big--must be less than %d", WORDSIZE_MAX+1);
				W_set = TRUE;
				break;
			case 'S':
				if (secondary) {
					if ((S2 = atol(a)) <= 0)
						bfatal(ERR_INVAL, "Bad value for S2");
					S2_set = TRUE;
					break;
				}
				if ((S = atol(a)) == 0)
					bfatal(ERR_INVAL, "Bad value for S");
				S_set = TRUE;
				break;
			case 'E':
				if (secondary) {
					if (sscanf(a, "%lg", &E2) != 1 || E2 > DEFAULT_E_MAX || E2 < 0.)
						bfatal(ERR_INVAL, "Bad value for E2");
					E2_set = 1;
					break;
				}
				if (sscanf(a, "%lg", &E) != 1 || E <= 0.0 || E > DEFAULT_E_MAX)
					bfatal(ERR_INVAL, "Bad value for E");
				E_set = TRUE;
				break;
			case 'T':
				if ((T = atoi(a)) < 1)
					bfatal(ERR_INVAL, "Bad value for T");
				T_set = TRUE;
				break;
			case 'X':
				if ((X = atoi(a)) < 1 || X > 100000)
					bfatal(ERR_INVAL, "Bad value for X");
				X_set = TRUE;
				minusX = -X;
				break;
			case 'Y':
				Y_set = TRUE;
				if ((Meff = atoi(a)) == 0)
					bfatal(ERR_INVAL, "Bad value for Y");
				break;
			case 'Z':
				if ((Neff = atoi(a)) == 0)
					bfatal(ERR_INVAL, "Bad value for Z");
				break;
			case 'C':
				if (isalpha(*a)) {
					display_gcodes();
					exit_code(ERR_NONE);
				}
				C = atoi(a);
				if (find_gcode(C) == NULL) {
					printf("Bad value for C\n");
					display_gcodes();
					exit_code(ERR_NONE);
				}
				break;
			case 'K':
				if (sscanf(a, "%lg", &K) != 1 || K < 0.)
					bfatal(ERR_INVAL, "Bad value for K");
				K_set = TRUE;
				break;
			case 'L':
				if (sscanf(a, "%lg", &Lambda) != 1 || Lambda <= 0.)
					bfatal(ERR_INVAL, "Bad value for L(ambda)");
				L_set = TRUE;
				break;
			case 'V':
				if (sscanf(a, "%ld", &V) != 1 || V < 0)
					bfatal(ERR_INVAL, "Bad value for V");
				V_set = TRUE;
				break;
			case 'B':
				if (sscanf(a, "%ld", &showblast) != 1 || showblast < 0)
					bfatal(ERR_INVAL, "Bad value for B");
				break;
			case 'H':
				showhist = atoi(a);
				break;
			default:
				usage();
		}
	}
	if (job_strands == 0)
		job_strands = (TOP_STRAND | BOTTOM_STRAND);

	if (showblast == 0 && showhist == 0 && V == 0)
		bfatal(ERR_INVAL, "All output is suppressed--no search will be performed");

	init_gcode(find_gcode(C), gcode, revgcode);

	/* read file of substitution weights */
	get_weights(M, AAID_MAX+1, aa_atob, &alphasize, &minchar, &maxchar, &achars,
					&order, &lows, &highs, &maxcost, &matrx);

	get_query(argv[2]);
	wrap(stdout, "Query=  ", qtitle, qtitlelen, 79, qnamelen+1);
	printf("        (%s letters", lu2s(q_len));
	Qlen = q_len + Qoffset;
	if (Qoffset != 0)
		printf(", offset %s", ld2s(Qoffset));
	printf(")\n");
 
	if (seqfilter(filtercmd, aa_btoa, aa_atob,
			qtitle, qtitlelen,
			(UcharPtr)q_seq, q_len,
			(UcharPtr)q_seq, (size_t PNTR)&q_len) != 0)
		bfatal(ERR_UNDEF, "Non-zero return code from sequence filter");
	
	putc('\n', stdout);
	initdb(argv[1]);

	(void) fflush(stdout);

	/*
	It would be more accurate to call lambdak() for every sequence in
	the database, but then the program would run much slower.
	Another suggestion is to call lambdak() before reporting
	the probability of an HSP containing part of a specific database
	sequence.  For the sake of speed, though, lambdak() is called only once.
	*/
	if (lambdak(lows, highs, q_seq, q_len, &efflen, &Lambda, &K, &H) != 0)
		bfatal(ERR_SCORING, "Could not calculate Lambda, K, and H");
	if (!Y_set)
		Meff = efflen;

	if (!X_set) {
		X = ceil(10./(Lambda/LN2)); /* score for 10 bits of information */
		minusX = -X;
	}

	if (!T_set) {
		T = DEFAULT_T;
		if (W == 3) {
			/* Ad hoc equation for PAM matrices with scale = ln(2)/2 only */
			T = ceil((5.376 + 1.3*log(H))/Lambda - 0.5);
			/*T = ceil((5.076 + 1.3*log(H))/Lambda - 0.5);*/
			/*T = ceil((2.7 + 2.741*H)/Lambda - 0.5);*/
		}
		if (W == 4) {
			/* Ad hoc equation for PAM matrices with scale = ln(2)/2 only */
			T = ceil((3.1 + 2.751*H)/Lambda - 0.5);
			/*T = ceil((2.8 + 2.751*H)/Lambda - 0.5);*/
		}
	}

	/* Determine the cutoff score, S */
	if (S_set && !E_set)
		E = 0.;
	/* only one reading frame is assumed to contain signal */
	cutoffs(&S, &E, &highs_exp, Lambda, K, H, Neff/CODON_LEN, Meff, gapdecayrate);

	/* Determine the secondary cutoff score, S2, to use */
	if (E2 == 0. && !S2_set)
		S2 = S;
	if ((E2_set && !S2_set && E2 <= 0.) || (S2_set && S2 > S))
		E2 = 0., S2 = S; /* S2 will be unused */
	else {
		if (S2_set && !E2_set)
			E2 = 0.;
		/* Assume an "average" protein's length is 300 a.a. */
		cutoffs(&S2, &E2, &highs_dummy, Lambda, K, H, 300, 300, 0.);
		/* Adjust S2 to be in line with S, as necessary */
		S2 = MAX(S2, 1);
		if (S2 > S)
			S2 = S;
		E2 = stoe(S2, Lambda, K, H, 300, 300);
	}

	score_hist_dim = etos(0.01, Lambda, K, H, Neff/CODON_LEN, Meff) + 1;
	score_hist = (unsigned long *)ckalloc0(sizeof(score_hist[0])*(score_hist_dim+1));
	poolp = HSPPoolNew();
	hl_pool = HitListPoolNew();

	/* The following relation is safe to use, because should the score
	ever reach S during hit extension, the entire diagonal will be
	rigorously searched ignoring X.
	*/
	X = MIN(X, S);
	minusX = -X;


	maxscore = fmaxscore(q_seq, q_len, maxcost);
	if (maxscore < S)
		bfatal(ERR_SCORING, "The maximum achievable score with this query sequence is only %d, whereas the cutoff score for reporting HSPs is %d.  You might lower the cutoff score, raise the expect value (E) to at least %ld, or use a different scoring scheme.  The maximum expect value this program will permit to be set is %lg.",
			maxscore, S,
			(long)gapdecay(1, stoe(maxscore, Lambda, K, H, Neff/CODON_LEN, Meff), gapdecayrate) + 1,
			DEFAULT_E_MAX);

	if (S < T)
		warning("The cutoff score for reporting HSPs, S = %d, is less than the neighborhood word score threshold, T = %d", S, T);

	sys_times(&t1);

	dp = bld_table_aa((DFA *)NULL, INT_MIN, q_seq, q_len, W, T, alphasize,
						&matrx, &maxcost, &order, &numwords, &numexact);
	end_table(dp);
	sys_times(&t2);

	RunWild("Searching", dbreccnt, dosearch);

#ifdef SIGXCPU
	/* Report results without interruption */
	signal(SIGXCPU, SIG_IGN);
#endif

	dfastatesize = dp->statesize;
	dfanstates = dp->nstates;
	dfasize = dfa_size(dp);
	dfaextent = dfa_extent(dp);
	dfa_destruct(dp);

	collate_hits();

	hasHSP = hitlist_count(firsthl);
	sort_hitlists(&firsthl, hitlist_cmp_func);
	nheaders = MIN(hasHSP, MAX(V, showblast));
	if (nheaders < hasHSP)
		hitlist_truncate(&firsthl, nheaders);
	get_headers(DATABASE_SEQTYPE, firsthl);
	sys_times(&t3);
	print_hist(score_hist_dim, score_hist, S, T, Lambda, K, H,
					Meff, Neff/CODON_LEN, nbr_seqs);
	print_headers(firsthl, 1.0, TRUE, 80);
	print_HSPs(firsthl, 1.0, print_t);
	print_parms();
	print_stats();
	ckwarnings();
	exit_code(ERR_NONE);
	/*NOTREACHED*/
}


print_parms()
{
	(void) printf("\n\nParameters:\n");
	if (filtercmd != NULL && filtercmd[0] != NULLB)
		(void) printf("  Query filter:  \"%s\"\n", filtercmd);
	(void) printf("  E = %#0.2lg, S = %ld (%.1lf bits),  E2 = %#0.2lg, S2 = %s\n",
			E, (long)S, (double)S*Lambda/LN2,
			E2, (S2 < S ? lu2s(S2) : "not used")
			);
	(void) printf("  W = %d, T = %ld (%.1lf bits), X = %ld (%.1lf bits)\n",
			W, (long)T, (double)T*Lambda/LN2, (long)X, (double)X*Lambda/LN2);
	(void) printf("  M = %s\n", M);
	(void) printf("  C = %d (%s genetic code)\n", C, find_gcode(C)->name);
	(void) printf("  H = %d, V = %d, B = %d\n", showhist, V, showblast);
	report_gapdecayrate();
	if (overlap_option != NULL)
		printf("  Overlap option:  %s\n", overlap_option);
	return ferror(stdout);
}

print_stats()
{
	(void) printf("\nStatistics:\n");
	(void) printf("  Lambda = %#0.3lg, Lambda/ln2 = %#0.3lg, K = %#0.3lg, H = %#0.3lg bits/position\n",
				Lambda, Lambda/LN2, K, H/LN2);
	(void) printf("  Expected/Observed high score = %s (%.1lf bits) / %s (%.1lf bits)\n",
			ld2s(highs_exp), ((double)highs_exp)*(Lambda/LN2),
			ld2s(highs_obs), ((double)highs_obs)*(Lambda/LN2)
			);
	if (S > highs_exp && highs_obs == 0)
		(void) printf("  NOTE:  the cutoff score is greater than the expected high score\n");
	if (highs_exp > maxscore)
		(void) printf("  NOTE:  expected high score is greater than the max. achievable score %d\n",
				maxscore);
	(void) printf("  # of letters in query:  %s aa", lu2s(q_len));
	if (Y_set)
		(void) printf("  (Y = %s)\n", lu2s(Meff));
	else
		if (Meff != q_len)
			(void) printf("  (Effective length = %s)\n", lu2s(Meff));
		else
			(void) putchar('\n');
	(void) printf("  # of neighborhood words in query:  %s\n", lu2s(numwords));
	(void) printf("  # of exact words scoring below T:  %s\n", lu2s(numexact));
	(void) wrap(stdout, "  Database:  ", dbtitle, -1, 79, 13);
	(void) printf("  # of letters in database:  %s bp", lu2s(totdblen));
	if (Neff != totdblen)
		(void) printf("  (Z = %s)\n", lu2s(Neff));
	else
		(void) putchar('\n');
	(void) printf("  # of word hits against database:  %s\n", lu2s((long)numhits));
	(void) printf("  # of failed hit extensions:  %s\n", lu2s((long)numfailed));
	(void) printf("  # of successful extensions:  %s\n", lu2s((long)totHSP+overcnt));
	(void) printf("  # of overlapping HSPs discarded:  %s\n",
				lu2s((long)overcnt));
	(void) printf("  # of HSPs reportable:  %s\n", lu2s((long)totHSP));
	(void) printf("  # of sequences in database:  %s\n", lu2s((long)nbr_seqs));
	(void) printf("  # of database sequences satisfying E:  %s\n",
				lu2s((long)hasHSP));

	if (numbogus && sfile != NULL)
		warning("*%d alignment%s included one or more X (unknown) amino acid codes where the database sequence contained nucleotide ambiguity codes and the translation of these codons could not be deduced unambiguously.",
			numbogus, PLURAL(numbogus));
	if (numbogus && sfile == NULL)
		warning("*%d alignment%s may be inaccurate because some database sequence(s) were known to contain non-ACGT(U) letters, while the original FASTA-format database file was unavailable to determine the locations of these letters.",
			numbogus, PLURAL(numbogus));

	(void) printf("No. of states in DFA:  %s (%s KB)\n",
			lu2s(dfanstates),
			lu2s(HOWMANY(dfanstates*dfastatesize, KBYTE)));
	(void) printf("Total size of DFA:  %s KB (%s KB)\n",
			lu2s(HOWMANY(dfasize, KBYTE)),
			lu2s(HOWMANY(dfaextent, KBYTE)));
	sys_showtime("Time to generate neighborhood", &t1, &t2, stdout);
	(void) printf("No. of processors used:  %d\n", numprocs);
	sys_showtime("Time to search database", &t2, &t3, stdout);
	sys_times(&t3);
	sys_showtime("Total cpu time", (Cputime_t *)NULL, &t3, stdout);
}

unsigned	hashsp;

static void
dosearch(userp, tp)
	Nlm_VoidPtr	userp;
	TaskBlkPtr	tp;
{
	int		frame, strand;
	int		hashsp_local = 0, rem_max_per_proc;
	Boolean	saved = FALSE;
	HSPPtr	hp;
	MPDEF

	rem_max_per_proc = REMEMBER_MAX / tp->proc_max;
#ifdef MPROC_AVAIL
	initmpc(tp->id, mp = &mpc[tp->id]);
#endif
	/* For each sequence in the database */
	while (TaskNext(tp) >= 0) {
		get_seq(MPC tp->task_cur + dbrecmin);
		MPPTR high_score = 0;
		if (MPPTR d_len >= CODON_LEN) {
			if (job_strands & TOP_STRAND) {
				strand = TOP_STRAND;
				for (frame = 1; frame <= CODON_LEN; ++frame) {
					translate(MPC frame);
					search_aa(MP);	/* search the database */
				}
			}
			if (job_strands & BOTTOM_STRAND) {
				strand = BOTTOM_STRAND;
				for (frame = -1; frame >= -CODON_LEN; --frame) {
					btranslate(MPC frame);
					search_aa(MP);	/* search the database */
				}
			}
		}

		if (MPPTR high_score < score_hist_dim)
			++MPPTR score_hist[MPPTR high_score];
		else
			++MPPTR score_hist[score_hist_dim];
		if (MPPTR high_score > MPPTR highs_obs)
			MPPTR highs_obs = MPPTR high_score;

		if (MPPTR curhl.hp != NULL) {
			sort_HSPs(&MPPTR curhl);
			if (consistency_flag) {
				consistn(&MPPTR curhl);
				consist_evals(&MPPTR curhl, Meff, MPPTR s_len, Neff/CODON_LEN);
			}
			else {
				pcnt(&MPPTR curhl);
				evals(&MPPTR curhl, Meff, MPPTR s_len, Neff/CODON_LEN);
			}
			(void) hitlist_prune_by_cutoff(&MPPTR curhl, MPPTR poolp, SCORET_EXPECT, E);
			if (MPPTR curhl.hp == NULL)
				continue;
			for (hp = MPPTR curhl.hp; hp != NULL; hp = hp->next)
				hp->P_val = etop(hp->expect);
			MPPTR curhl.seq_nbr = tp->task_cur + dbrecmin;
			if (++hashsp_local + hashsp > rem_max_per_proc) {
				M_LOCK;
				if (!saved) {
					saved = TRUE;
					hashsp += hashsp_local - 1;
					hashsp_local = 0;
				}
				if (++hashsp > REMEMBER_MAX)
					bfatal(ERR_REMEMBER,
					"More than %u sequences have HSPs to remember; increase REMEMBER_MAX and recompile.",
					REMEMBER_MAX);
				M_UNLOCK;
			}
			hitlist_save(MPPTR hl_pool, &MPPTR firsthl, &MPPTR curhl);
			MPPTR totHSP += MPPTR curhl.hspcnt;
			MPPTR curhl.hp = NULL;
			MPPTR curhl.hspcnt = 0;
		}
	}
	M_LOCK;
	hashsp += hashsp_local;
	M_UNLOCK;
}


#ifdef MPROC_AVAIL
/* Initialize the processor context pointed to by mp */
initmpc(id, mp)
	int		id;
	MPContextPtr	mp;
{
	if ((mp->id = id) == 0) {
		mp->_dlev = _dlev;
		mp->diag_level = diag_level;
		mp->diag_top = mp->diag_stack = diag_top;
		mp->s_seq0 = s_seq0;
		mp->s_seq1 = s_seq1;
		mp->s_seq2 = s_seq2;
		mp->score_hist = score_hist;
		mp->poolp = poolp;
		mp->hl_pool = hl_pool;
	}else {
		long	len;
		len = maxlen/CODON_LEN + q_len + 1;
		mp->_dlev = (DIAG_TYPE *)ckalloc0(sizeof(*_dlev)*len);
		mp->diag_level = mp->_dlev + q_len;
		mp->diag_top = mp->diag_stack = (DIAG_PTR)ckalloc(sizeof(*diag_top)*len);
		mp->s_seq0 = (char *)ckalloc(maxlen+1+CODON_LEN) + 1;
		mp->s_seq1 = mp->s_seq0 + maxlen/CODON_LEN + 1;
		mp->s_seq2 = mp->s_seq1 + maxlen/CODON_LEN + 1;
		mp->s_seq0[-1] = mp->s_seq1[-1] = mp->s_seq2[-1] = NULLB;
		mp->score_hist
			= (unsigned long *)ckalloc0(sizeof(mp->score_hist[0])*(score_hist_dim+1));
		mp->poolp = HSPPoolNew();
		mp->hl_pool = HitListPoolNew();
	}
}
#endif


display_gcodes()
{
	register int	i;

	printf("The available genetic codes are:\n");
	for (i=0; i<DIM(gcodes); ++i)
		printf("  %d. %s\n", gcodes[i].id, gcodes[i].name);
	printf("Specify a particular code by its number (%d is the default)\n",
		DEFAULT_C);
}

translate(MPC theframe)
	MPDEF
	int		theframe;
{
	MPPTR frame = theframe;
	if (theframe == 1) {
		register int	i;
		register unsigned char	ch;
		register unsigned char	*sp1, *sp2, *sp3;
		register unsigned char	*cp;
		register unsigned char	*cend;
		unsigned char	*spray[CODON_LEN];
		int		j;

		sp1 = (unsigned char *) (MPPTR s_seq = MPPTR s_seq0);
		sp2 = (unsigned char *) MPPTR s_seq1;
		sp3 = (unsigned char *) MPPTR s_seq2;

		cp = (unsigned char *) MPPTR d_seq;
		cend = cp + MPPTR d_len/4;
		ch = *cp;
		i = (ch>>6);
		i <<= 2;
		i |= (ch>>4)&3;
		j = MPPTR d_len%4;

#define XBASE0(p)	i<<=2, i&=0x3f, i|=(ch>>6), *p++ =gcode[i]
#define XBASE1(p)	i<<=2, i&=0x3f, i|=(ch>>4)&3, *p++ =gcode[i]
#define XBASE2(p)	i<<=2, i&=0x3f, i|=(ch>>2)&3, *p++ =gcode[i]
#define XBASE3(p)	i<<=2, i&=0x3f, i|=ch&3, *p++ =gcode[i]

		if (cp < cend) {
			for (;;) {
				XBASE2(sp1);
				XBASE3(sp2);
				if (++cp >= cend)
					goto Break1;
				ch = *cp;
				XBASE0(sp3);
				XBASE1(sp1);
				XBASE2(sp2);
				XBASE3(sp3);
				if (++cp >= cend)
					goto Break2;
				ch = *cp;
				XBASE0(sp1);
				XBASE1(sp2);
				XBASE2(sp3);
				XBASE3(sp1);
				if (++cp >= cend)
					goto Break3;
				ch = *cp;
				XBASE0(sp2);
				XBASE1(sp3);
			}
Break1:
			spray[0] = sp3;
			spray[1] = sp1;
			spray[2] = sp2;
			goto Break;
Break2:
			spray[0] = sp1;
			spray[1] = sp2;
			spray[2] = sp3;
			goto Break;
Break3:
			spray[0] = sp2;
			spray[1] = sp3;
			spray[2] = sp1;
Break:
			if (j == 0)
				goto Terminate;
			ch = *cp;
			XBASE0(spray[0]);
			if (j > 1)
				XBASE1(spray[1]);
			if (j > 2)
				XBASE2(spray[2]);
		}
		else {
			if (j > 2)
				XBASE2(sp1);
			spray[0] = sp1;
			spray[1] = sp2;
			spray[2] = sp3;
		}

Terminate:
		*spray[0] = *spray[1] = *spray[2] = NULLB;
		MPPTR s_len = MPPTR d_len/CODON_LEN;
		return;
	}
	if (theframe == 2) {
		MPPTR s_seq = MPPTR s_seq1;
		MPPTR s_len = (MPPTR d_len-1)/CODON_LEN;
		if (MPPTR s_len < 0)
			MPPTR s_len = 0;
	}
	else { /* theframe == 3 */
		MPPTR s_seq = MPPTR s_seq2;
		MPPTR s_len = (MPPTR d_len-2)/CODON_LEN;
		if (MPPTR s_len < 0)
			MPPTR s_len = 0;
	}
}

btranslate(MPC theframe)
	MPDEF
	int		theframe;
{
	MPPTR frame = theframe;
	if (theframe == -1) {
		register int	i;
		register unsigned char	ch;
		register char	*sp1, *sp2, *sp3;
		register char	*cp = (char *)MPPTR d_seq;
		register char	*cbeg;
		char	*spray[CODON_LEN];
		int		j;

		sp1 = spray[0] = MPPTR s_seq = MPPTR s_seq0;
		sp2 = spray[1] = MPPTR s_seq1;
		sp3 = spray[2] = MPPTR s_seq2;
		if (MPPTR d_len == 0) {
			*sp1 = *sp2 = *sp3 = NULLB;
			MPPTR s_len = 0;
			return;
		}

		cbeg = MPPTR d_seq;
		cp = MPPTR d_seq + (MPPTR d_len+3)/4 - 1;
		j = MPPTR d_len%4;
		/* skip over pad bits */
		ch = (unsigned char)*cp;
		if (j == 0)
			j = 4;
		else
			ch >>= 2*(4-j);

#define XBASE(p)	i<<=2, i&=0x3c, i|=ch&3, ch>>=2, *p++ =revgcode[i]

		/* Finish off the terminal byte */
		i = ch&3;
		if (j > 1) {
			ch >>= 2;
			i <<= 2;
			i |= ch&3;
			ch >>= 2;
			if (j > 2) {
				XBASE(sp1);
				if (j == 4) {
					XBASE(sp2);
				}
			}
		}

		if (cp > cbeg) {
			ch = (unsigned char)*--cp;
			/* Start work on the penultimate byte */
			switch (j) {
				case 1:
					i <<= 2;
					i |= ch&3;
					ch >>= 2;
					XBASE(sp1);
					spray[0] = sp2;
					spray[1] = sp3;
					spray[2] = sp1;
					break;
				case 2:
					XBASE(sp1);
					XBASE(sp2);
					spray[0] = sp3;
					spray[1] = sp1;
					spray[2] = sp2;
					break;
				case 3:
					XBASE(sp2);
					XBASE(sp3);
					spray[0] = sp1;
					spray[1] = sp2;
					spray[2] = sp3;
					break;
				case 4:
					XBASE(sp3);
					XBASE(sp1);
					spray[0] = sp2;
					spray[1] = sp3;
					spray[2] = sp1;
					break;
			}
			sp1 = spray[0];
			sp2 = spray[1];
			sp3 = spray[2];
			/* Finish the penultimate byte and process all further bytes */
			for (;;) {
				XBASE(sp1);
				XBASE(sp2);
				if (--cp < cbeg)
					break;
				ch = (unsigned char)*cp;
				XBASE(sp3);
				XBASE(sp1);
				XBASE(sp2);
				XBASE(sp3);
				if (--cp < cbeg)
					break;
				ch = (unsigned char)*cp;
				XBASE(sp1);
				XBASE(sp2);
				XBASE(sp3);
				XBASE(sp1);
				if (--cp < cbeg)
					break;
				ch = (unsigned char)*cp;
				XBASE(sp2);
				XBASE(sp3);
			}
		}

		*sp1 = *sp2 = *sp3 = NULLB;
		MPPTR s_len = MPPTR d_len/CODON_LEN;
		return;
	}
	if (theframe == -2) {
		MPPTR s_seq = MPPTR s_seq1;
		MPPTR s_len = (MPPTR d_len-1)/CODON_LEN;
		if (MPPTR s_len < 0)
			MPPTR s_len = 0;
	}
	else { /* theframe == -3 */
		MPPTR s_seq = MPPTR s_seq2;
		MPPTR s_len = (MPPTR d_len-2)/CODON_LEN;
		if (MPPTR s_len < 0)
			MPPTR s_len = 0;
	}
}

/* get_query - read the query sequence */
get_query(filename)
	char *filename;
{
	char	*base, *bp;
	size_t	basemax;
	FILE	*fp;
	register char *p, ch;
		
	fp = ckopen(filename, "r", 1);

	base = ckalloc(basemax = 10*KBYTE);
	bp = base + 1;

	while (vfgets(&base, &basemax, &bp, fp) != NULL) {
		if (*bp == '>') {	/* Optional header/title line */
			if (bp != base+1)/* Second sequence encountered */
				break;
			p = str_chr(bp, '\n');
			if (p != NULL)
				*p = NULLB;
			qtitle = str_dup(bp+1);
			qtitlelen = strlen(qtitle);
			for (p=qtitle; (ch = *p) && !isspace(ch); ++p)
				;
			*p = NULLB;
			qname = str_dup(qtitle);
			*p = ch;
			qnamelen = strlen(qname);
			continue;
		}
		p = bp;
		while ((ch = *p++) != NULLB) {
			if (aa_atob[ch] <= AAID_MAX) {
				if (islower(ch))
					ch = toupper(ch);
				*bp++ = ch;
			}
			else
				if (aa_atob[ch] == AAID_NAR) {
					aa_atob[ch] = AAID_IGNORE;
warning("Invalid amino acid code ``%c'' encountered in query sequence will be skipped.", ch);
				}
		}
		if (bp - base > QUERYLEN_MAX-1)
			bfatal(ERR_QUERYLEN, "query sequence exceeds %d characters in length",
				QUERYLEN_MAX-1);
			
	}
	(void) fclose(fp);
	_q_seq = base;
	_q_seq[0] = NULLB;
	q_seq = base+1;
	q_len = bp - q_seq;
	if (q_len < W)
		bfatal(ERR_QUERYLEN, "query sequence is shorter than the word length, W=%d.", W);
	*bp = NULLB;

	(void) ckseqtype(QUERY_SEQTYPE, q_seq, q_len);

	p = q_seq;
	while ((ch = *p) != NULLB)
		*p++ = aa_atob[ch];

	qq_seq = q_seq;
	ss_seq = ckalloc(q_len+1);
	return 0;
}

/* -----------------------   read the compressed database  -------------------*/
/* UNIX-specific input routine; database must be processed by SETDB */

static void
initdb(dbname)
	char	*dbname;
{
	long	many, len;
	long	cleaned;
	int		once = 0;
	char	*dirname = "";
	long	i;

RETRY:
	if (*dirname != NULLB)
		(void) sprintf(dbfname, "%s%s%s", dirname, DIRDELIMSTR, dbname);
	else
		SAFENCPY(dbfname, dbname, sizeof(dbfname));
	(void) sprintf(name, "%s%s", dbfname, TABLE_EXT);
	tfile = mfil_open(name, "r", MFIL_OPT_ALL);
	if (tfile == NULL) {
		if ((dirname = getenv("BLASTDB")) == NULL)
			dirname = BLASTDB;
		if (++once == 1)
			goto RETRY;
		bfatal(ERR_DBASE, "Couldn't find database file ``%s''.  Make sure you have specified the correct database name and requested a database of the correct type (nucleotide) for use with this program.", name);
	}

	/* It's now okay if the original FASTA format file can't be found */
	sfile = ckopen(dbfname, "r", 0);

	(void) mfil_dup_long(tfile, &dbtype, 1);
	if (dbtype != DB_TYPE_NUC)
		bfatal(ERR_DBASE, "``%s'' is not a nucleotide sequence database", dbname);
	(void) mfil_dup_long(tfile, &dbformat, 1);
	if (dbformat != NTFORMAT)
		bfatal(ERR_DBASE, "Incorrect database format");

	if (mfil_dup_long(tfile, &i, 1) != 1)
		bfatal(ERR_DBASE, "Corrupted database file:  %s", name);
	if ((dbtitle = mfil_get(tfile, i)) == NULL)
		bfatal(ERR_DBASE, "Corrupted database file:  %s", name);
	if (i%4 != 0) {                                      
		i = 4 - (i%4);
		mfil_seek(tfile, i, SEEK_CUR);
	}

	(void) mfil_dup_long(tfile, (long *)&line_len, 1);
	(void) mfil_dup_long(tfile, (long *)&nbr_seqs, 1);
	(void) mfil_dup_long(tfile, (long *)&maxlen, 1);
	(void) mfil_dup_long(tfile, (long *)&totdblen, 1);
	(void) mfil_dup_long(tfile, (long *)&c_len, 1);
	if (mfil_dup_long(tfile, (long *)&clean_count, 1) != 1)
		bfatal(ERR_DBASE, "corrupted database tab file");

	/* Skip over the list of cleaned 8-mers */
	(void) mfil_seek(tfile, clean_count * BO_LONG_SIZE, SEEK_CUR);

	if (dbtitle[0] == NULLB)
		dbtitle = str_dup(dbname);
	(void) wrap(stdout, "Database:  ", dbtitle, -1, 79, 11);
	(void) printf("           %s sequences; %s total letters.\n",
			ld2s(nbr_seqs), ld2s(totdblen));                       

	if (nbr_seqs <= 0 || totdblen <= 0)
		bfatal(ERR_DBASE, "Nothing in the database to search!?");
 
	many = nbr_seqs + 1;
	cseq_beg = (unsigned long *)mfil_get_long(tfile, many);
	seq_beg = (unsigned long *)mfil_get_long(tfile, many);
	header_beg = (unsigned long *)mfil_get_long(tfile, many);
	ambiguity_ray = (char *)mfil_get(tfile, nbr_seqs/CHAR_BIT + 1);
	if (cseq_beg == NULL || seq_beg == NULL || header_beg == NULL ||
			ambiguity_ray == NULL)
		bfatal(ERR_DBASE, "Corrupted database tab file");

	(void) sprintf(name, "%s%s", dbfname, NT_HEADER_EXT);
	hfile = mfil_open(name, "r", MFIL_OPT_ALL);
	if (hfile == NULL)
		bfatal(ERR_FOPEN, "Couldn't open database file %s", name);

	(void) sprintf(name, "%s%s", dbfname, NT_SEARCHSEQ_EXT);
	/* Attach a shared memory segment for the .csq file (if one exists) */
	cfile = mfil_open(name, "r", MFIL_OPT_ALL);
	if (cfile == NULL)
		bfatal(ERR_FOPEN, "Couldn't open database file %s", name);
	if ((c_seq = (PACK_TYPE *)MFILE_SP(cfile)) == NULL) {
#ifdef MPROC_AVAIL
		if (c_len > CDBLEN_MAX)
			bfatal(ERR_DBASE, "database is too large; raise CDBLEN_MAX and recompile");
		fflush(stdout);
		c_seq = (PACK_TYPE *)ckalloc(c_len);
		dbbuf = (CharPtr)c_seq;
		if (mfil_read((CharPtr)c_seq, c_len, 1, cfile) != 1)
			bfatal(ERR_DBASE, "Cannot read compressed database sequence file %s", name);
#else
		d_seq = (char *)ckalloc(maxlen/(CHAR_BIT/NBPN) + 2 + NSENTINELS);
		(void) mfil_seek(cfile, cseq_beg[0]/(CHAR_BIT/NBPN), SEEK_SET);
#endif
	}

	c_len -= NSENTINELS; /* Adjust length for the NT_MAGIC_BYTEs at end */

	/* Establish pointers to start of translations in 3 reading frames */
	s_seq0 = (char *)ckalloc(maxlen+1+CODON_LEN);
	s_seq0 += 1;
	s_seq1 = s_seq0 + maxlen/CODON_LEN + 1;
	s_seq2 = s_seq1 + maxlen/CODON_LEN + 1;
	/* Set NULL prefix sentinel bytes for the 3 reading frames */
	s_seq0[-1] = s_seq1[-1] = s_seq2[-1] = NULLB;

	len = maxlen/CODON_LEN + q_len + 1;
	_dlev = (DIAG_TYPE *)ckalloc0(sizeof(*_dlev)*len);
	diag_level = _dlev + q_len;

	diag_top = diag_stack = (DIAG_PTR)ckalloc(sizeof(*diag_top)*len);

	if (Neff == 0)
		Neff = totdblen;

	dbrecmin = MAX(dbrecmin, 0);
	dbrecmin = MIN(dbrecmin, nbr_seqs);
	dbrecmax = (dbrecmax > 0 ? MIN(dbrecmax, nbr_seqs) : nbr_seqs);
	dbreccnt = dbrecmax - dbrecmin;
	if (dbreccnt < nbr_seqs) {
		nbr_seqs = dbreccnt;
		printf("           Subset of database from %ld to %ld.\n",
			dbrecmin+1, dbrecmax);
	}
}

static int
get_seq(MPC i)
	MPDEF
	register unsigned long	i;
{
	MPPTR c_len = cseq_beg[i+1]/(CHAR_BIT/NBPN) - cseq_beg[i]/(CHAR_BIT/NBPN);

	if (c_seq != NULL)
		MPPTR d_seq = (char *)c_seq + cseq_beg[i]/(CHAR_BIT/NBPN);
	else {
		if (mfil_read(d_seq, c_len, 1, cfile) != 1)
			bfatal(ERR_DBASE, "error reading compressed database sequence");
	}
	MPPTR current = i;
	MPPTR c_len -= NSENTINELS;
	MPPTR c_last = MPPTR c_len - 1;
	MPPTR c_lastp = MPPTR c_seq + MPPTR c_last;

	/* Calculate the uncompressed length */
	MPPTR d_len = (MPPTR c_len)*(CHAR_BIT/NBPN);
	MPPTR c_pad = cseq_beg[i] & ((CHAR_BIT/NBPN) - 1);
	if (MPPTR c_pad != 0)
		MPPTR d_len -= ((CHAR_BIT/NBPN) - MPPTR c_pad);
	return 0;
}

/*****************************/
#include "searcha.inc"
/*****************************/

/* ----------------------------  parameter calculations  ---------------------*/
/* Version of lambdak() specifically for a.a. alphabet used by BLASTP */
int
lambdak(lows, highs, q_seq, q_len, efflen, Lambda, K, H)
	long	lows, highs;
	unsigned long	q_len;
	long	PNTR efflen;
	register char	*q_seq;
	double	*Lambda, *K, *H;
{
	double	fq1[AAID_MAX+1];
	unsigned long	iq1[AAID_MAX+1];
	double	*prob;
	Score_t	range;
	unsigned long	qlen;
	register Score_t	ui;
	register int	i, j;

	if (lows >= 0 || highs <= 0)
		return 1;

	/* Tally the frequency of occurrence of each letter in the query */
	for (i=0; i<DIM(iq1); ++i)
		iq1[i]=0;
	for (ui=0, qlen = q_len; ui<q_len; ++ui) {
		if (q_seq[ui] != aa_atob[UNKNOWN_AA_CHR])
			++iq1[q_seq[ui]];
		else
			--qlen;
	}
	*efflen = qlen;

	/* Convert frequencies to probabilities */
	for (i=AAID_MIN; i<=AAID_MAX; ++i)
		fq1[i] = (double)iq1[i] / qlen;

	/* Calculate probability of each score */
	range = highs - lows;
	prob = (double *)ckalloc(sizeof(*prob)*(range+1));
	for (ui=0; ui<=range; ++ui)
		prob[ui] = 0.;

	for (i=minchar; i<=maxchar; ++i)
		for (j=AAID_MIN; j<DIM(aafq); ++j) {
			if (matrx[i][j] > SCORE_MIN/WORDSIZE_MAX)
				prob[matrx[i][j]-lows] += fq1[i]*aafq[j];
		}

	i = karlin(lows, highs, prob, !L_set, Lambda, !K_set, K, TRUE, H);
	free((char *)prob);
	if (i != 0) {
		*K = *Lambda = -1.;
		return i;
	}
	return 0;
}

static int
print_t(hlp)
	HitListPtr	hlp;
{
	long	nseqlen;
	Seqid_t	current;
	int		i, npos, nlen;
	int		ambiguous, prevframe = 0;
	HSPPtr	hp;
	CharPtr	scoreflag;
	register char	*cp, ch;

	current = hlp->seq_nbr;
	ambiguous = ambiguity_ray[current/CHAR_BIT] & (1 << (current%CHAR_BIT));
	if (sfile == NULL && ambiguous)
		scoreflag = "*";
	else
		scoreflag = "";
	nseqlen = ((cseq_beg[current+1]>>NBPN) - (cseq_beg[current]>>NBPN) - NSENTINELS) << NBPN;
	i = cseq_beg[current] & BITMASK(NBPN);
	if (i != 0)
		nseqlen -= (4-i);

	cp = hlp->header;
	while ((ch = *cp) != NULLB && !isspace(ch))
		++cp;
	while ((ch = *cp) != NULLB && isspace(ch))
		++cp;
	i = cp - hlp->header + 1;
	i = MIN(i, 12);
	putchar('\n');
	putchar('\n');
	(void) wrap(stdout, ">", hlp->header, hlp->header_len, 79, i);
	(void) printf("%*sLength = %s\n", i, "", ld2s((long)nseqlen));

	for (hp = hlp->hp; hp != NULL; hp = hp->next) {
		nlen = hp->len * CODON_LEN;
		if (hp->frame > 0)
			npos = hp->s_pos*CODON_LEN + hp->frame - 1;
		else
			npos = nseqlen - hp->s_pos*CODON_LEN + hp->frame - nlen + 1;

		if (ambiguous && sfile != NULL) {
			get_seq_from_fasta(s_seq0, npos, nlen, sfile, seq_beg[current],
									line_len);
			if (hp->frame > 0)
				trans2(s_seq0, ss_seq, hp->len);
			else
				rtrans2(s_seq0, ss_seq, hp->len);
		}
		else {
			get_seq_from_cseq(s_seq0, npos, nlen, cfile, c_seq,
								cseq_beg[current]/(CHAR_BIT/NBPN));
			if (hp->frame > 0)
				trans2b(s_seq0, ss_seq, hp->len);
			else
				rtrans2b(s_seq0, ss_seq, hp->len);
		}

		if (SIGN(prevframe) != SIGN(hp->frame)) {
			if (hp->frame > 0)
				printf("\n  Plus");
			else
				printf("\n  Minus");
			printf(" Strand HSPs:\n");
		}

		if (hsp_print(stdout, hp, scoreflag, (UcharPtr)qq_seq + hp->q_pos, (UcharPtr)ss_seq, matrx, aa_btoa, aa_btoa, Qlen, nseqlen, nox_coord, nox_coord, apo_to_nt1, apo_to_nt3, "Frame", str_frame, Lambda) != 0)
			return 1;
		prevframe = hp->frame;
	}
	return 0;
}

static void
trans2(ntq, aaq, n)
	register char	*ntq;	/* input nucleotide sequence */
	register char	*aaq;	/* output amino acid sequence */
	int		n;			/* length of translated protein */
{
	register char	*aaqmax = aaq+n;
	register unsigned	i, j, k;

	while (aaq < aaqmax) {
		i = nt_atob[*ntq++];
		j = nt_atob[*ntq++];
		k = nt_atob[*ntq++];
		*aaq++ = codon2aa(gcode, i, j, k);
	}
	/* NOT NULL TERMINATED! */
}

static void
rtrans2(ntq, aaq, n)
	register char	*ntq, *aaq;
	int		n;
{
	register char	*aaqmax = aaq+n;
	register unsigned	i, j, k;

	ntq += CODON_LEN*n;
	while (aaq < aaqmax) {
		i = nt_atob[*--ntq];
		j = nt_atob[*--ntq];
		k = nt_atob[*--ntq];
		*aaq++ = codon2aa(revgcode, i, j, k);
	}
	/* NOT NULL TERMINATED! */
}

static void
trans2b(ntq, aaq, n)
	register char	*ntq;	/* input nucleotide sequence */
	register char	*aaq;	/* output amino acid sequence */
	int		n;			/* length of translated protein */
{
	register char	*aaqmax = aaq+n;
	register unsigned	i, j, k;

	while (aaq < aaqmax) {
		i = *ntq++;
		j = *ntq++;
		k = *ntq++;
		*aaq++ = codon2aa(gcode, i, j, k);
	}
	/* NOT NULL TERMINATED! */
}

static void
rtrans2b(ntq, aaq, n)
	register char	*ntq, *aaq;
	int		n;
{
	register char	*aaqmax = aaq + n;
	register unsigned	i, j, k;

	ntq += CODON_LEN*n;
	while (aaq < aaqmax) {
		i = *--ntq;
		j = *--ntq;
		k = *--ntq;
		*aaq++ = codon2aa(revgcode, i, j, k);
	}
	/* NOT NULL TERMINATED! */
}

usage()
{
	fflush(stdout);
	fflush(stderr);
	fprintf(stderr, "\nUsage:\n\n");
	fprintf(stderr,
		"    %s database query_sequence [options]\n\n", module);
	fprintf(stderr,
		"Valid %s options:  E, S, E2, S2, W, T, X, M, C, Y, Z, K, L, H, V and B\n",
		module);
	fprintf(stderr, "\t-qoffset #\n");
	fprintf(stderr, "\t-dbrecmin #\n");
	fprintf(stderr, "\t-dbrecmax #\n");
	fprintf(stderr, "\t-gi (gi identifier suppression)\n");
	fprintf(stderr, "\ttop\n\tbottom\n\t-overlap\n\t-overlap2\n\t-gapdecayrate #\n\t-filter filteroption\n\t-echofilter\n\t-consistency\n");
	fprintf(stderr, "\t-sort_by_pvalue\n\t-sort_by_count\n\t-sort_by_highscore\n\t-sort_by_totalscore\n");
	exit_code(ERR_INVAL);
}
